unit ahGLRandomHDS;
{
    This unit provides tools and objects to generate random Height Data Sources
  that can be used with tGLTerrainRenderer. General properties are defined in
  tGLBaseRandomHDS, but the main object is tGLCustomRandomHDS,
  which defines all the basic functionalities; however, it is an abstract class.
  So far, it has only one descendent, tGLFractalHDS, which implements the fractal
  middle-point displacement algorithm (aka plasma, aka diamond-square).
    The actual algorithms are independent functions called by the objects so they
  can also be used in other contexts. Basically, only the BuildHeightField
  method has to be overriden, and properties
  particular to the algorithm added (see tGLFractalHDS implementation). The
  BuildHeightField should contain a call to the algorithm function (or the algorithm
  itself, and MUST set the following fields: fSize, fMinHeight, fMaxHeight and
  fRangeHeight.

    Landscape generation consists in the following steps:
  1 Generate height field
  2 Modify it through erosion, sea surface, etc.
  3 Compute light and shadows
  4 Build the texture and assign it to a material created for this purpose

    The above classes generate isolated landscapes. They can be tiled in an
  infinite landscape through tGLTiledRndLandscape. The function of this class
  is to maintain a list of landscapes (called thereafter "landtiles"), to build
  and free them when needed.
    The tGLFractalArchipelago is an example of such a landscape generating an
  infinite landscape made of fractal islands.

    Although this structure may appear complex, the programmer just need to
  instanciate a tGLFractalArchipelago and to set its properties to get it running
  transparently. See the FractalLandscape and FractalArchipelago demos to see
  how to use these objects and what the various properties mean.

    Additional comments can be found in the code in the particular procedures.

    These components can be freely used. Please send me any improvement you made
  to this code ( Alexandre.Hirzel@nat.unibe.ch ). So far, you have to declare and
  create this component manually in your code and link it to a tGLTerrainRenderer.
  If you know how to make a registered component, please do it.

                                Alexandre Hirzel

  History:
  05.07.2003:
    - tGLTiledRndLandscape released
    - tGLFractalArchipelago released
    - Water plane and associated parameters (thanks to Eric Grange)
    - Other classes modified accordingly
    - HeightData generated as smallint in place of single
  25.06.2003:
    - Texture drawed for each tile independently
    - Steps property added
    - tGLTiledRndLandscape to deal with infinite landscapes
    - Further hierarchisation
  22.06.2003:
    - Normal map
    - Changed from relative to absolute Z-coordinates
    - ClearHeightField, predefined height-field.
    - Bug fixed in ClearLightMap
  21.06.2003:
    - First implementation
}

interface
         
uses Windows,Forms,Classes,Graphics,SysUtils,Contnrs,jpeg,
     GLScene,Geometry,GLHeightData,GLMisc,GLTerrainRenderer,GLTexture,
      OpenGL12;

type
  tSeaErosion=record
    Enabled		:boolean;
    BeachHeight    :single;
  end;

  tRainErosion=record
    Enabled		:boolean;
    ErosionRate   	:single;
    DepositRate 	:single;
  end;

  tLifeErosion=record
    Enabled		:boolean;
    Robustness	:single;
  end;

  tFractionErosion=record
    Enabled		:boolean;
    Slope		:single;
  end;

  tLandTileInfo=record
    x,z		:integer;          // Coordinates of the landtile. Used to generate the seed
    State	:THeightDataState; // Preparation status of the landtile
  end;

  tSteps=record
    Enabled	:boolean;
    Count	:integer;
  end;

  tMapOfSingle= array of array of single;
  tMapOfVector= array of array of TVector;

  tGLBaseRandomHDS= class;

  {Function type to use for topography-based texture}
  tOnDrawTexture=function(const Sender:tGLBaseRandomHDS; x,y:integer; z:double;
                          Normal:tVector):tColorVector of object;

  tSingleClamp=procedure(var x,y:single)of object;
  tIntegerClamp=procedure(var x,y:integer)of object;

  tGLBaseRandomHDS= class(THeightDataSource)
  {This class introduces all the basic properties of random landscape. No method
   implemented though. It is used as a descendant for
   - tGLCustomRandomLandscape: one tile landscape (cyclic or not)
   - tGLTiledRndLandscape: "infinite" landscapes (grids of tGLCustomRandomLandscape)}
  private
    FSteps: tSteps;
    FLandCover: boolean;
    procedure SetOnDrawTexture(const Value: tOnDrawTexture);
    procedure SetSteps(const Value: tSteps);
    procedure SetLandCover(const Value: boolean);
  protected
    FSeed: integer;
    FSize: integer;
    FMaterialName: string;
    FLighting: boolean;
    FLightDirection:tVector;
    FTerrainRenderer: TGLTerrainRenderer;
    FLightColor: tColorVector;
    FShadows: boolean;
    FSea: boolean;
    FSeaLevel: single;
    FAmbientLight: single;
    FErosionByRain: tRainErosion;
    FErosionBySea: tSeaErosion;
    FTaskProgress: integer;
    FTextureScale: integer;
    FErosionByLife: tLifeErosion;
    FErosionByFraction: tFractionErosion;
    FLightSmoothing: boolean;
    FCyclic: boolean;
    FSeaTransparency: single;
    FPrimerLandscape: boolean;
    FLandTileInfo: tLandTileInfo;
    FOnDrawTexture: tOnDrawTexture;
    function OnDrawTextureDefault(const Sender:tGLBaseRandomHDS; x,y:integer; z:double;
                          Normal:tVector):tColorVector;
    procedure SetSeed(const Value: integer);
    procedure SetMaterialName(const Value: string);
    procedure SetLighting(const Value: boolean);
    procedure SetLightDirection(Value: TVector);
    procedure SetTerrainRenderer(const Value: TGLTerrainRenderer);virtual;abstract;
    procedure SetLightColor(const Value: tColorVector);
    procedure SetShadows(const Value: boolean);
    procedure SetSea(const Value: boolean);
    procedure SetSeaLevel(const Value: single);
    procedure SetAmbientLight(const Value: single);
    procedure SetErosionByRain(const Value: tRainErosion);
    procedure SetErosionBySea(const Value: tSeaErosion);
    procedure SetTextureScale(const Value: integer);
    procedure SetErosionByLife(const Value: tLifeErosion);
    procedure SetErosionByFraction(const Value: tFractionErosion);
    procedure SetLightSmoothing(const Value: boolean);
    procedure SetSeaTransparency(const Value: single);
    procedure SetPrimerLandscape(const Value: boolean);
    function GetSeaLevel: single;
    function GetSeaTransparency: single;
    procedure SetLandTileInfo(const Value: tLandTileInfo);
    procedure SetCyclic(const Value: boolean); virtual;abstract;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    {Usually white, but you can generate sunset ambiance by setting it to red,
    for example.}
    property LightColor:tColorVector read FLightColor write SetLightColor;
    {Light is parallel (sun light)}
    property LightDirection:TVector read FLightDirection write SetLightDirection;
    {This function must be supplied by the user. Here he/she can define which
    colour to use depending on coordinates, elevation and normal. This provides
    a great flexibility. If no function is supplied (OnDrawTexture=nil), a default
    texture function is used (very basic, just blue and green).}
    property OnDrawTexture:tOnDrawTexture read FOnDrawTexture write SetOnDrawTexture;
  published
    property AmbientLight:single read FAmbientLight write SetAmbientLight;
    {If true, the landscape can be tiled to itself seamlessly.
    If false, the landscape is an isolated square.}
    property Cyclic:boolean read FCyclic write SetCyclic;
    {Erosion parameters. See associated record types}
    property ErosionByFraction:tFractionErosion read FErosionByFraction write SetErosionByFraction;
    property ErosionByLife:tLifeErosion read FErosionByLife write SetErosionByLife;
    property ErosionByRain:tRainErosion read FErosionByRain write SetErosionByRain;
    property ErosionBySea:tSeaErosion read FErosionBySea write SetErosionBySea;
    property LandCover:boolean read FLandCover write SetLandCover;
    {Enable or disable all lighting effects}
    property Lighting:boolean read FLighting write SetLighting;
    {True by default. You can gain a little speed by disabling it.}
    property LightSmoothing:boolean read FLightSmoothing write SetLightSmoothing;
    {Not used}
    property MaterialName:string read FMaterialName write SetMaterialName;
    {If true, the height-field will not be emptied and generation will take the
     existing heights to shape the new landscape}
    property PrimerLandscape:boolean read FPrimerLandscape write SetPrimerLandscape;
    {Enable the sea surface truncation}
    property Sea:boolean read FSea write SetSea;
    {Sea level}
    property SeaLevel:single read GetSeaLevel write SetSeaLevel;
    {Depth at which the sea bottom becomes invisible. See DoSea for more information}
    property SeaTransparency:single read GetSeaTransparency write SetSeaTransparency;
    {Seed used by the random generator. Each seed generate a different
    reproductible landscape.}
    property Seed:integer read FSeed write SetSeed;
    {Enable shadow casting. May take some time for large Depth.}
    property Shadows:boolean read FShadows write SetShadows;
    property Steps:tSteps read FSteps write SetSteps;
    {TerrainRenderer used to render the HDS.}
    property TerrainRenderer:TGLTerrainRenderer read FTerrainRenderer write SetTerrainRenderer;
    {Defines how many texture pixels are drawn per height-field cell. The larger
    this number the better the quality of the resulting image, but it takes a
    more time to compute. Good results are got between 1 and 5.}
    property TextureScale:integer read FTextureScale write SetTextureScale;
  end;

  tGLCustomRandomHDS=class(tGLBaseRandomHDS)
  {Base structure for all random landscape objects. It can't be used directly
  since its BuildHeightField procedure is abstract. Use one of its descendants
  instead.}
  private
    fSlave		:boolean;
    fTextureCreated:boolean;
    fMaxHeight	:single;
    fMinHeight	:single;
    fRangeHeight	:single;
    FTask: string;
    fSingleConstrain:tSingleClamp;
    fIntegerConstrain:tIntegerClamp;
    FKeepNormals: boolean;
    procedure CreateTexture;
    function GetHeight(x, y: integer): single;
    procedure SetHeight(x, y: integer; const Value: single);
    procedure SetKeepNormals(const Value: boolean);
  protected
    procedure SetTerrainRenderer(const Value: TGLTerrainRenderer);override;
    procedure SetCyclic(const Value: boolean);override;
    procedure BoundaryClamp(var x,y:single);overload;
    procedure BoundaryClamp(var x,y:integer);overload;
    procedure CyclicClamp(var x,y:single);overload;
    procedure CyclicClamp(var x,y:integer);overload;
    {tTerrainRenderer event handler}
    procedure GetTerrainBounds(var l, t, r, b: Single);
    {This procedure MUST be called by the descendent of tGLBaseRandomHDS}
    procedure SetSize(const aSize:integer);
    {tTerrainRender event handler}
    procedure StartPreparingData(heightData : THeightData); override;
  public
    fHeight		:tMapOfSingle;
    fLightMap		:tMapOfSingle;
    fNormal		:tMapOfVector;
    // Upper bounds of the tile
    function BoundaryX:integer;
    function BoundaryZ:integer;
    // Generate the heightfield array, based on the topographical properties
    procedure BuildHeightField;overload;virtual;abstract;
    {Provide an automated way to build a landscape. However, a greater control can
     be achieved by calling the various procedures manually (they are public methods)
     as one gets a sligthly different result depending on the sequence of erosion
     and sea steps.}
    procedure BuildLandscape;
  { - Compute the light effects
    - Compute the casted shadows
    - Perform a basic smoothing if TextureScale>1}
    procedure BuildLightMap;overload;
    procedure BuildLightMap(const aLightDirection:tVector);overload;
    {Normals are needed for lighting and slope-based textures}
    procedure BuildNormals;
    {For every pixel of the texture, computes slope and interpolated height and
     sends these information to a user-supplied function (OnDrawTexture), whose
     result is a tColorVector. If no OnDrawTexture is supplied, a basic default
     texture will be used.}
    procedure BuildTexture;
    {Fill the heightfield with "Empty" values (-999)}
    procedure ClearHeightField;
    {Fill the light map with 1}
    procedure ClearLightMap;
    {Constrain x,y to be in the boundaries of the height field array. This is
    done in two way depending on the kind of landscape:
    Cyclic landscapes: 		mod
    Non-cyclic landscape:	clamp}
    procedure ConstrainCoordinates(var x,y:single); overload;
    procedure ConstrainCoordinates(var x,y:integer);overload;
    constructor Create(AOwner: TComponent); override;
    constructor CreateAsSlave(AOwner:TComponent);
    destructor Destroy; override;
    {Enforces an identical height on the opposing edges of the landscape}
    procedure DoCyclicBoundaries;
    {Not yet implemented}
    procedure DoErosionByFraction;
    {Just a smoothing. Should be done last as it improves the look of other
     erosion effects. Too much biological erosion can ruin erase their results
     though. Some tweaking may be needed}
    procedure DoErosionByLife;
    {Creare sharp valleys and canyons. If DepositRate>0, it will also fill the
    low pools, producing flat "lakes" and "ponds" }
    procedure DoErosionByRain;
    {Create a beach and a cliff around the islands}
    procedure DoErosionBySea;
    {Cut all elevations lower than sea level. If Transparency>0, the sea surface
    will not be flat, but a slight elevation change (unperceptible in 3D view)
    allow to fake transparency in the OnDrawTexture.}
    procedure DoSea;
    {Discretise the heigthfield in a chosen number of steps}
    procedure DoSteps;
    {x and y are range-checked and constrained into the array. This slows down
    computation. If you don't need to range-check (this is mainly useful in
    cyclic landscapes when you need a seamless joint), call fHeigth instead
    (this is a protected field, therefore only accessible from tGLFractalHDS
    descendents.}
    property Height[x,y:integer]:single read GetHeight write SetHeight; // Range checked
    {A specific implementation of THeightDataSource.InterpolatedHeight}
    function Interpolate(x,y:single):single;
    {Keep the array of normals for future use}
    property KeepNormals:boolean read FKeepNormals write SetKeepNormals;
    {Property used by tGLTiledRndLandscape to know where the landtile is located
     and other parameters. See tLandTileInfo}
    property LandTileInfo:tLandTileInfo read FLandTileInfo write SetLandTileInfo;
    {Range checking}
    function PointInMap(const x,y:single):boolean; overload;
    function PointInMap(const x,y:integer):boolean; overload;
    {Store the minimum and maximum elevations}
    property MaxHeight:single read fMaxHeight;
    property MinHeight:single read fMinHeight;
    {Vector normal to the terrain at the position}
    function Normal(Position:tVector):tVector;
    {Max height - min height}
    property RangeHeight:single read fRangeHeight;
    {Scale of the Terrain Renderer. They are set so as giving a identical
    vertical/horitontal ratio with any size. Therefore, Scale.X=Scale.Y=1 and
    only Scale.Z varies. If you want to increase the landscape scale, the best way
    would be to place the Terrain Renderer in a DummyCube and rescale it.}
    function Scale:TGLCoordinates;
    {Size of the square height array. With the middle-point algorithm, it is always
    Size = 2^N+1. In a cyclic landscape, the last row and columns are identical
    to the first. }
    property Size:integer read FSize;
    {A height rescaled between 0 and 1000 for}
    function StandardisedHeight(const x,y:integer):single;
    {When long computations are running, this property contains the operation
    beeing processed.}
    property Task:string read FTask;
    {A value between 0 and 100 indicating the percentage of completion}
    property TaskProgress:integer read FTaskProgress;
    {Use these boundaries with non-cyclic landscapes to constrain camera movements.}
    function XMoveBoundary:single;
    function ZMoveBoundary:single;
  published
    property Cyclic:boolean read FCyclic write SetCyclic;
  end;//class


  tGLFractalHDS=class(tGLCustomRandomHDS)
  {Random landscape based on the middle-point displacement algorithm}
  private
    fAmplitude: integer;
    fDepth: integer;
    fRoughness: single;
    procedure SetAmplitude(const Value: integer);
    procedure SetDepth(const Value: integer);
    procedure SetRoughness(const Value: single);
  public
    procedure BuildHeightField;overload;override;
    procedure BuildHeightField(const aDepth,aSeed,aAmplitude:integer);overload;
    constructor Create(AOwner:tComponent); override;
  published
    {Proportional to the difference between highest and lowest altitude.}
    property Amplitude:integer read FAmplitude write SetAmplitude;
    {Number of levels in the fractal process. Depth defines the size of the
    landscape: Size = 2^Depth+1 . Good results are got with Depth>=6. Above 10
    the landscape takes a lot of time to be generated.}
    property Depth:integer read FDepth write SetDepth;
    {The lower this parameter, the smoother the landscape. Takes value between 0 and 1}
    property Roughness:single read FRoughness write SetRoughness;
  end;// class

  //tMapOfLandscapes	:array of array of tGLBaseRandomHDS;

  tLandTile=tGLCustomRandomHDS;
  tRelativeCoordinate=record
    dx,dz	:integer
  end;

  tOnCreateLandTile=procedure(x,z,Seed:integer;var aLandscape:tLandTile) of object;
  tIsDefaultTile=function(x,z:integer):boolean of object;

  tGLTiledRndLandscape=class(tGLBaseRandomHDS)
  private
    fLandTileComputing:boolean; // Is a landtile being computed?
    FExtentX: integer;
    FExtentZ: integer;
    FExtentXhalf: integer;
    FExtentZhalf: integer;
    fLandTileSize:integer;
    fSingleConstrain:tSingleClamp;
    fIntegerConstrain:tIntegerClamp;
    FTerrainRenderer: TGLTerrainRenderer;
    FCamera: TGLCamera;
    fOnCreateLandTile: tOnCreateLandTile;
    fOnCreateDefaultTile: TStartPreparingDataEvent;
    FIsDefaultTile: tIsDefaultTile;
    FSeed: integer;
    fBaseSeed:integer;
    fComputedLandTile:tLandTile;
    FLandTileCapacity: integer;
    FGenerationRadius: integer;
    FLandTileDensity: single;
    procedure fDefaultOnCreateDefaultTile(HeightData : THeightData);
    function fDefaultIsDefaultTile(x,z:integer):boolean;
    procedure SetExtentX(const Value: integer);
    procedure SetExtentZ(const Value: integer);
    procedure SetOnCreateLandTile(const Value: tOnCreateLandTile);
    procedure SetCamera(const Value: TGLCamera);
    procedure SetIsDefaultTile(const Value: tIsDefaultTile);
    procedure SetSeed(const Value: integer);
    procedure SetOnCreateDefaultTile(const Value: TStartPreparingDataEvent);
    function GetTask: string;
    function GetTaskProgress: integer;
    procedure SetLandTileCapacity(const Value: integer);
    procedure SetGenerationRadius(const Value: integer);
    procedure SetLandTileDensity(const Value: single);
  protected
    fGenRadius	:array of tRelativeCoordinate;
    fOldCamX	:integer;
    fOldCamZ	:integer;
    fMapUpdating	:boolean;
    fLandTiles	:tComponentList;
    procedure BoundaryClamp(var x,z:single);overload;
    procedure BoundaryClamp(var x,z:integer);overload;
    procedure ComputeLandTile(const aX,aZ: integer;var NewLandTile:tLandTile);virtual;
    procedure CyclicClamp(var x,z:single);overload;
    procedure CyclicClamp(var x,z:integer);overload;
    {tTerrainRenderer event handler}
    procedure GetTerrainBounds(var l, t, r, b: Single);
    function LandTileSeed(x,z:integer):integer;
    property OnCreateDefaultTile:TStartPreparingDataEvent read fOnCreateDefaultTile write SetOnCreateDefaultTile;
    procedure SetCyclic(const Value: boolean); override;
    {This procedure MUST be called by the descendent of tGLRandomArchipelago}
    procedure SetSize(const aSize:integer);
    function fSortLandscapes(Item1, Item2: Pointer): Integer;
    //procedure PrepareLandTileData(HeightData:tHeightData; LandTile:tLandTile);
    {tTerrainRender event handler}
    procedure StartPreparingData(HeightData : tHeightData); override;
    procedure SetTerrainRenderer(const Value: TGLTerrainRenderer);override;
  public
    procedure ApplyLighting(var aLandTile: tLandTile);
    procedure ApplyTexture(var aLandTile: tLandTile);
    procedure ApplyTopography(var aLandTile: tLandTile);
    procedure CameraPosition(var TileX, TileZ:integer);
    procedure CleanUp;
    {Constrain x,y to be in the boundaries of the height field array. This is
    done in two way depending on the kind of landscape:
    Cyclic landscapes: 		mod
    Non-cyclic landscape:	clamp}
    procedure ConstrainCoordinates(var x,z:single); overload;
    procedure ConstrainCoordinates(var x,z:integer);overload;
    constructor Create(AOwner: TComponent); override;
    destructor Destroy;override;
    {Compute the landtile containing (x,z)}
    procedure FindLandTile(const x,z:single; var TileX,TileZ:integer);
    {Build the first landtile and position the camera. Must be called first.}
    procedure Initialize(const ax,az:single);virtual;
    {User-supplied function determining if this landtile will be built by the
    OnCreateDefaultTile or if a landscape will be generated.}
    property IsDefaultTile:tIsDefaultTile read FIsDefaultTile write SetIsDefaultTile;
    {Number of landtile in memory}
    function LandtileCount:integer;
    {Size of a landtile. Must be a power of two}
    property LandTileSize:integer read fLandTileSize;
    {User-specified event handler containing the particular code for tile generation}
    property OnCreateLandTile:tOnCreateLandTile read FOnCreateLandTile write SetOnCreateLandTile;
    {When long computations are running, this property contains the operation
    beeing processed.}
    property Task:string read GetTask;
    {A value between 0 and 100 indicating the percentage of completion}
    property TaskProgress:integer read GetTaskProgress;
    {Distance between two landtiles}
    function TileDistance(const x1,z1,x2,z2:integer):single;
    {Square of distance between two landtiles. Use this function to compare
    two distances.}
    function TileDistanceSquared(const x1,z1,x2,z2:integer):integer;
    {This procedure check which landtiles must be generated or destroyed as a
     function of camera position. This is let to the descendent classes.}
    procedure Update;
    property MapUpdating:boolean read fMapUpdating;
    {Use these boundaries with non-cyclic landscapes to constrain camera movements.}
    function XMoveBoundary:single;
    function ZMoveBoundary:single;
  published
    property Camera:TGLCamera read FCamera write SetCamera;
    property Cyclic:boolean read FCyclic write SetCyclic;
    {Dimensions of the "infinite" landscape. Can be set very high. These parameters
     have neither memory nor speed consequence. They are mainly used to compute
     a unique seed for each landtile}
    property ExtentX:integer read FExtentX write SetExtentX;
    property ExtentZ:integer read FExtentZ write SetExtentZ;
    {Distance at which a new landtile begin to be built. Increasing this value
     allows for a higher camera speed but it will also increase the memory requirements.}
    property GenerationRadius:integer read FGenerationRadius write SetGenerationRadius;
    {Number of landtile to keep in memory. Should not be modified.}
    property LandTileCapacity:integer read FLandTileCapacity write SetLandTileCapacity;
    {Probability that a given landtile is non-default}
    property LandTileDensity:single read FLandTileDensity write SetLandTileDensity;
    {Base seed for the entire archipelago}
    property Seed:integer read FSeed write SetSeed;
    {Terrain renderer linked to the HDS. Must be set just after creation.}
    property TerrainRenderer:TGLTerrainRenderer read FTerrainRenderer write SetTerrainRenderer;
  end;//class

  tGLFractalArchipelago=class(tGLTiledRndLandscape)
  private
    FDepth: integer;
    FRoughnessMax: single;
    FRoughnessMin: single;
    FAmplitudeMin: integer;
    FAmplitudeMax: integer;
    FSeaDynamic: boolean;
    FSeaMaterialName: string;
    FWaveAmplitude: single;
    FWaveSpeed: single;
    function GetIslandDensity:single;
    {PostRender event handler drawing a static water plane between islands
    Code borrowed from Eric's Archipelago GLScene advanced demo}
    procedure fPostRenderSeaStatic(var rci: TRenderContextInfo; const heightDatas: TList);
    {Sea with waves.
    Borrowed from Eric's Archipelago GLScene advanced demo}
    procedure fPostRenderSeaDynamic(var rci: TRenderContextInfo; const heightDatas: TList);
    procedure SetIslandDensity(const Value:single);
    procedure SetDepth(const Value: integer);
    procedure SetRoughnessMax(const Value: single);
    procedure SetRoughnessMin(const Value: single);
    procedure SetAmplitudeMax(const Value: integer);
    procedure SetAmplitudeMin(const Value: integer);
    procedure SetSeaDynamic(const Value: boolean);
    procedure SetSeaMaterialName(const Value: string);
    procedure SetWaveAmplitude(const Value: single);
    procedure SetWaveSpeed(const Value: single);
  protected
    procedure SetTerrainRenderer(const Value: TGLTerrainRenderer);override;
    procedure fOnCreateLandTile(aX,aZ,aSeed:integer;var aLandscape:tLandTile);
    procedure fOnCreateDefaultTile(HeightData : THeightData);
  public
    procedure ComputeLandTile(const aX,aZ: integer;var NewLandTile:tLandTile); override;
    constructor Create(AOwner: TComponent); override;
  published
    {Ranges for the amplitude parameter in the fractal algorithm}
    property AmplitudeMax:integer read FAmplitudeMax write SetAmplitudeMax;
    property AmplitudeMin:integer read FAmplitudeMin write SetAmplitudeMin;
    {Depth of the fractal algorithm}
    property Depth:integer read FDepth write SetDepth;
    {A wrapper for LandtileDensity. This is the probabilty for a landtile to
    contain an island.}
    property IslandDensity:single read GetIslandDensity write SetIslandDensity;
    {Ranges for the roughness parameter in the fractal algorithm}
    property RoughnessMax:single read FRoughnessMax write SetRoughnessMax;
    property RoughnessMin:single read FRoughnessMin write SetRoughnessMin;
    {If true, the sea will show dynamic waves. Slow.}
    property SeaDynamic:boolean read FSeaDynamic write SetSeaDynamic;
    {Reference to a material in the TerrainRenderer's material library. This
    material will be used to drape the water plane.}
    property SeaMaterialName:string read FSeaMaterialName write SetSeaMaterialName;
    {Size of the waves}
    property WaveAmplitude:single read FWaveAmplitude write SetWaveAmplitude;
    property WaveSpeed:single read FWaveSpeed write SetWaveSpeed;
  end;//class

  {Texture functions}
  function LoadJPGtexture(const JpgName:string):tBitmap;
  function NoisyColor(const Color:tColor;const Noise:single=0.05): TColorVector;
  function TextureGreen(const x,y:integer):tColorVector;
  function TextureBlue(const x,y:integer):tColorVector;
  function TextureSand(const x, y: integer): tColorVector;
  function TextureBrownSoil(const x, y: integer): tColorVector;
  function TextureDarkGreen(const x, y: integer): tColorVector;
  function TextureDarkGray(const x, y: integer): tColorVector;
  function TextureWhite(const x, y: integer): tColorVector;

  {Random HDS functions}
  procedure FractalMiddlePointHDS(const aDepth, aSeed,aAmplitude: integer;
                             const aRoughness:single; aCyclic:boolean;
                             var z:tMapOfSingle; var MinZ,MaxZ:single);
  procedure InitializeRandomGenerator(const Seed:integer);

  {Landscape primers}
  procedure PrimerNull(var z:tMapOfSingle);
  procedure PrimerIsland(LowZ,HighZ:single;var z:tMapOfSingle);

const
  VerticalScalingFactor=128;

implementation

const {Neighbourhood vectors and weight}
  NeighX	:array[0..8] of integer=(-1, 0, 1, 1, 1, 0,-1,-1, 0);
  NeighY	:array[0..8] of integer=(-1,-1,-1, 0, 1, 1, 1, 0, 0);
  NeighW :array[0..8] of single=(1/1.4142,1,1/1.4142,1,1/1.4142,1,1/1.4142,1,2);
  SumWeights=4/1.4142 + 4 + 2;
  Empty:single=-999;
  VSF=VerticalScalingFactor;

var
  rhdsStartTime		:cardinal;
  rhdsLandscapeCounter	:cardinal=0;
  //Counter	:tTickCounter;

function LoadJPGtexture(const JpgName:string):tBitmap;
var
  Jpg	:TJPEGImage;
begin
  Result:=TBitmap.Create;
  Jpg:=TJPEGImage.Create;
  Jpg.LoadFromFile(JpgName);
  Result.Assign(Jpg);
  Jpg.Free;
end;

function NoisyColor(const Color:tColor;const Noise:single=0.05): TColorVector;
var
  r	:single;
begin
  Result:=ConvertWinColor(Color);
  r:=random*Noise;
  AddVector(Result,r);
end;

function TextureSand(const x, y: integer): tColorVector;
begin
  Result:=NoisyColor($0071D8FF);
end;

function TextureBrownSoil(const x, y: integer): tColorVector;
begin
  Result:=NoisyColor($00008BBF);
end;

function TextureDarkGreen(const x, y: integer): tColorVector;
begin
  Result:=NoisyColor($00004000);
end;

function TextureDarkGray(const x, y: integer): tColorVector;
begin
  Result:=NoisyColor(clDkGray);
end;

function TextureWhite(const x, y: integer): tColorVector;
begin
  Result:=NoisyColor(clWhite);
end;

function TextureBlue(const x, y: integer): tColorVector;
begin
  Result:=NoisyColor(clBlue);
end;

function TextureGreen(const x, y: integer): tColorVector;
begin
  Result:=NoisyColor(clGreen);
end;

procedure InitializeRandomGenerator(const Seed:integer);
var
  i	:integer;
begin
  RandSeed:=Seed;
  for i:=1 to 50 do random;  // Pre-heat the generator 
end;

{ tGLBaseRandomHDS }

constructor tGLBaseRandomHDS.Create(AOwner: TComponent);
begin
  inherited;
  Inc(rhdsLandscapeCounter);
  Name:=Format('RandomLandscape%d',[rhdsLandscapeCounter]);
  fLightColor:=VectorMake(1,1,1);
  fLightDirection:=VectorMake(-1,0,-1);
  fAmbientLight:=0.5;
  fTextureScale:=1;
  fMaterialName:='';
  FLighting:=True;
  FLightSmoothing:=True;
  Cyclic:=True;
  FSeed:=RandSeed;
  FSeaLevel:=0.0;
  FErosionBySea.BeachHeight:=0.01;
  FErosionBySea.Enabled:=False;
  FErosionByRain.Enabled:=True;
  FErosionByRain.ErosionRate:=0.5;
  FErosionByRain.DepositRate:=FErosionByRain.ErosionRate;
  FErosionByLife.Enabled:=True;
  FErosionByLife.Robustness:=1;
  fLandTileInfo.State:=hdsNone;
end;

destructor tGLBaseRandomHDS.Destroy;
begin
  inherited;
end;

function tGLBaseRandomHDS.GetSeaLevel: single;
begin
  Result:=FSeaLevel/VSF; // factor used in tTerrainRender
end;

function tGLBaseRandomHDS.GetSeaTransparency: single;
begin
  Result:=FSeaTransparency/VSF; // factor used in tTerrainRender
end;

function tGLBaseRandomHDS.OnDrawTextureDefault(
  const Sender: tGLBaseRandomHDS; x, y: integer; z: double;
  Normal: tVector): tColorVector;
begin
  if z>Sender.SeaLevel*VSF then Result:=TextureGreen(x,y) else Result:=TextureBlue(x,y);
end;

procedure tGLBaseRandomHDS.SetAmbientLight(const Value: single);
begin
  FAmbientLight := Value;
end;

procedure tGLBaseRandomHDS.SetErosionByFraction(
  const Value: tFractionErosion);
begin
  FErosionByFraction := Value;
end;

procedure tGLBaseRandomHDS.SetErosionByLife(const Value: tLifeErosion);
begin
  FErosionByLife := Value;
end;

procedure tGLBaseRandomHDS.SetErosionByRain(const Value: tRainErosion);
begin
  FErosionByRain := Value;
end;

procedure tGLBaseRandomHDS.SetErosionBySea(const Value: tSeaErosion);
begin
  FErosionBySea := Value;
end;

procedure tGLBaseRandomHDS.SetLandCover(const Value: boolean);
begin
  FLandCover := Value;
end;

procedure tGLBaseRandomHDS.SetLandTileInfo(const Value: tLandTileInfo);
begin
  FLandTileInfo := Value;
end;

procedure tGLBaseRandomHDS.SetLightColor(const Value: tColorVector);
begin
  FLightColor := Value;
end;

procedure tGLBaseRandomHDS.SetLightDirection(Value: TVector);
var
  v	:tVector;
begin
  v:=Value;
  NormalizeVector(v);
  FLightDirection:=Value;
end;

procedure tGLBaseRandomHDS.SetLighting(const Value: boolean);
begin
  FLighting := Value;
end;

procedure tGLBaseRandomHDS.SetLightSmoothing(const Value: boolean);
begin
  FLightSmoothing := Value;
end;

procedure tGLBaseRandomHDS.SetMaterialName(const Value: string);
begin
  FMaterialName := Value;
end;

procedure tGLBaseRandomHDS.SetOnDrawTexture(const Value: tOnDrawTexture);
begin
  if @Value<>nil then FOnDrawTexture := Value
  else FOnDrawTexture:=OnDrawTextureDefault; // Basic texture event
end;

procedure tGLBaseRandomHDS.SetPrimerLandscape(const Value: boolean);
begin
  FPrimerLandscape := Value;
end;

procedure tGLBaseRandomHDS.SetSea(const Value: boolean);
begin
  FSea := Value;
end;

procedure tGLBaseRandomHDS.SetSeaLevel(const Value: single);
begin
  FSeaLevel := Value*VSF; // factor used in tTerrainRender
end;

procedure tGLBaseRandomHDS.SetSeaTransparency(const Value: single);
begin
  FSeaTransparency := Value*VSF; // factor used in tTerrainRender
end;

procedure tGLBaseRandomHDS.SetSeed(const Value: integer);
begin
  FSeed := Value;
end;

procedure tGLBaseRandomHDS.SetShadows(const Value: boolean);
begin
  FShadows := Value;
end;

procedure tGLBaseRandomHDS.SetSteps(const Value: tSteps);
begin
  FSteps := Value;
end;

procedure tGLBaseRandomHDS.SetTextureScale(const Value: integer);
begin
  FTextureScale := Value;
end;

{ tGLCustomRandomHDS }

procedure tGLCustomRandomHDS.BoundaryClamp(var x, y: single);
begin
  ClampValue(x,0,fSize);
  ClampValue(y,0,fSize);
end;

procedure tGLCustomRandomHDS.BoundaryClamp(var x, y: integer);
begin
  if x<0 then x:=0 else if x>fSize then x:=fSize;
  if y<0 then y:=0 else if y>fSize then y:=fSize;
end;

function tGLCustomRandomHDS.BoundaryX: integer;
begin
  Result:=Round(FSize*Scale.X);
end;

function tGLCustomRandomHDS.BoundaryZ: integer;
begin
  Result:=Round(FSize*Scale.Z);
end;

procedure tGLCustomRandomHDS.BuildLandscape;
{Provides an automated way to build a landscape. However, a greater control can
 be achieved by calling the various procedures manually (they are public methods)
 as one gets a sligthly different result depending on the sequence of erosion
 and sea steps.}
begin
  fTask:='Landscape generation';
  FTaskProgress:=0;

  {Empty all height-field cells}
  if not FPrimerLandscape then ClearHeightField;

  {Build the basic fractal height field. It is mandatory and must always be
   called first.}
  BuildHeightField;

  {Various operations that reshape the height field. These procedures may be
   called in any order, although the one proposed here is the most natural.
   These procedures are optional}
  if FErosionByRain.Enabled then DoErosionByRain;
  if FErosionByLife.Enabled then DoErosionByLife;
  if FErosionBySea.Enabled then DoErosionBySea;
  if FSteps.Enabled then DoSteps;

  {Doing sea first would speeds up the following processes
  but the result would be slightly less realistic. In
  particular with transparency, you can have a nice effect
  of submarine valleys prolungating land canyons.
  This procedure is optional}
  if FSea then DoSea;

  if FCyclic then DoCyclicBoundaries; // Ensures a seamless fit

  {Compute a normal for each vertex. Used by BuildLightMap and BuildTexture}
  if fLandCover then BuildNormals;

  {Add light effects. Either BuildLightMap or ClearLigthMap must be called.
   Used by BuildTexture.}
  if fLighting and LandCover then BuildLightMap else ClearLightMap;

  {Builds the actual texture. If it is not used, the terrain will be textured
  with its Material, if defined.}
  if fLandCover then BuildTexture;

  {Free memory. If you need often to recompute texture, you may want to keep
  one or both maps, providing the heightfield or the light source have not changed.}
  if not FKeepNormals then fNormal:=nil;
  fLightMap:=nil;

  fTask:=' Updating terrain renderer';
  FTaskProgress:=0;
  Application.ProcessMessages;

  MarkDirty;  //Tells the HDS that changes have been made (don't forget it or you'll get strange things)
end;

procedure tGLCustomRandomHDS.BuildLightMap;
{ - Compute the light effects
  - Compute the casted shadows
  - Perform a basic smoothing if TextureScale>1}
var
  i,j,k,m,n	:integer;
  x,y		:single;
  t			:single;
  v1,v2		:TVector;
  L			:tVector;
  Shade		:single;
begin
  if fSize=0 then exit;
  fTask:='Light-map computation';
  fTaskProgress:=0;
  SetLength(fLightMap,(fSize+1)*TextureScale,(fSize+1)*TextureScale);
  L:=fLightDirection;
  NormalizeVector(L);
  NegateVector(L);

  {Compute lighting}
  for i:=0 to fSize do begin
    fTaskProgress:=Round(i/fSize*100);
    for j:=0 to fSize do begin
      Application.ProcessMessages;
      Shade:=abs(VectorDotProduct(fNormal[i,j],L));
      ClampValue(Shade,0);
      for k:=i*TextureScale to (i+1)*TextureScale-1 do
        for n:=j*TextureScale to (j+1)*TextureScale-1 do
          fLightMap[k,n]:=Shade;
    end;//for
  end;//for i

  {Shadows}
  if FShadows then begin
    fTask:='Shadow casting';
    fTaskProgress:=0;
    L[0]:=L[0]*Scale.X;
    L[1]:=L[1]*VSF/Scale.Y;
    L[2]:=L[2]*Scale.Z;
    for j:=0 to fSize do begin
      fTaskProgress:=Round(j/fSize*100);
      for i:=0 to fSize do begin
        if fLightMap[i*TextureScale,j*TextureScale]>0 then begin // Don't look for shadow if the point is already shadowed
          v1:=VectorMake(i,fHeight[i,j],j); // Starting point
          for k:=2 to Round(fSize*1.4) do begin // Quick and dirty ray-casting
            v2:=VectorCombine(v1,L,1,k); // Casts a ray in direction of the sun
            x:=Round(v2[0]);
            y:=Round(v2[2]);
            if Interpolate(x,y)>v2[1] then begin
              Application.ProcessMessages;
              for m:=i*TextureScale to (i+1)*TextureScale-1 do
                for n:=j*TextureScale to (j+1)*TextureScale-1 do
                  fLightMap[m,n]:=0;
              break; // Shadow caster found. No need to continue
            end;//if
          end;//for k
        end;//if
      end;//for j
    end;//for i
  end;//if

  {Smoothing}
  if fLightSmoothing then begin
    fTask:='Light-map smoothing';
    fTaskProgress:=0;
    for m:=1 to TextureScale-1 do begin
      fTaskProgress:=Round(m/TextureScale*100);
      for j:=1 to High(fLightMap)-1 do begin
        for i:=1 to High(fLightMap)-1 do begin
          Application.ProcessMessages;
          t:=0;
          for k:= 0 to 8 do begin
            t:=t+fLightMap[i+NeighX[k],j+NeighY[k]]*NeighW[k];
          end;//for k
          fLightMap[i,j]:=t/SumWeights;
        end;//for j
      end;//for i
    end;//for m
  end;//if
end;

procedure tGLCustomRandomHDS.BuildLightMap(const aLightDirection: tVector);
begin
  FLightDirection:=aLightDirection;
  BuildLightMap;
end;

procedure tGLCustomRandomHDS.BuildNormals;
var
  i,j		:integer;
  z0		:single;
  v1,v2		:TVector;
  n1		:tVector;
  Normal		:tVector;
begin
  fTask:='Normal computation';

  for i:=0 to fSize do begin
    fTaskProgress:=Round(i/fSize*100);
    for j:=0 to fSize do begin
      Application.ProcessMessages;
      z0:=fHeight[i,j];
      Normal:=NullHmgVector;
      MakeVector(v1,Scale.X,(Height[i+1,j]-z0)*Scale.Y/VSF,0);
      MakeVector(v2,0,(Height[i,j+1]-z0)*Scale.Y/VSF,Scale.Z);
      Normal:=VectorCrossProduct(v2,v1);
      NormalizeVector(Normal);
      MakeVector(v1,-Scale.X,(Height[i-1,j]-z0)*Scale.Y/VSF,0);
      MakeVector(v2,0,(Height[i,j+1]-z0)*Scale.Y/VSF,Scale.Z);
      n1:=VectorCrossProduct(v1,v2);
      NormalizeVector(n1);
      Normal:=VectorAdd(Normal,n1);
      MakeVector(v1,-Scale.X,(Height[i-1,j]-z0)*Scale.Y/VSF,0);
      MakeVector(v2,0,(Height[i,j-1]-z0)*Scale.Y/VSF,-Scale.Z);
      n1:=VectorCrossProduct(v2,v1);
      NormalizeVector(n1);
      Normal:=VectorAdd(Normal,n1);
      MakeVector(v1,Scale.X,(Height[i+1,j]-z0)*Scale.Y/VSF,0);
      MakeVector(v2,0,(Height[i,j-1]-z0)*Scale.Y/VSF,-Scale.Z);
      n1:=VectorCrossProduct(v1,v2);
      NormalizeVector(n1);
      Normal:=VectorAdd(Normal,n1);
      fNormal[i,j]:=VectorScale(Normal,0.25); // Average of the 4 adjacent normals
    end;//for j
  end;//for i
end;

procedure tGLCustomRandomHDS.BuildTexture;
type
  pRGBTripleArray = ^TRGBTripleArray;
  tRGBTripleArray = array [word] of TRGBTriple;

var
  Bmp	:array of array of tBitmap;
  x0,y0	:integer;
  xx,yy	:integer;
  x,y	:integer;
  nbTiles:integer;
  Side	:integer;
  meancol:tColor;
  Line	:pRGBTripleArray;

  function MeanColor(color1,color2 : TColor) : TColor;
  var
    r1,g1,b1 : Byte;
    r2,g2,b2 : Byte;
  begin
    r1:=(color1 and $000000FF);
    g1:=((color1 and $0000FF00) shr 8);
    b1:=((color1 and $00FF0000) shr 16);
    r2:=(color2 and $000000FF);
    g2:=((color2 and $0000FF00) shr 8);
    b2:=((color2 and $00FF0000) shr 16);
    Result:=RGB((r1+r2)div 2,(g1+g2)div 2,(b1+b2)div 2);
  end;

  procedure MakeRGBTriple(const Color:tColorVector;var RGBTriple:tRGBTriple);
  begin
    with RGBTriple do begin
      rgbtRed:= Round(Color[0]*255);
      rgbtGreen:= Round(Color[1]*255);
      rgbtBlue:= Round(Color[2]*255);
    end;//with
  end;

  function ComputePixel(const x,y:integer):tRGBTriple;
  var
    i,j	:integer;
    Shade	:tColorVector;
    Cover	:tColorVector;
    z		:double;
  begin
    i:=(x0+x) div TextureScale;
    j:=(y0+y) div TextureScale;
    z:=Interpolate((x0+x)/TextureScale,(y0+y)/TextureScale);
    Application.ProcessMessages;
    {Cover:=OnDrawTexture(Self,FLandTileInfo.x*fSize+x0+x,
                              FLandTileInfo.z*fSize+y0+y,z,fNormal[i,j]);}
    Cover:=OnDrawTexture(Self,x0+x,y0+y,z,fNormal[i,j]);
    Application.ProcessMessages;
    Shade:=VectorScale(fLightColor,fLightMap[x0+x,y0+y]);
    Application.ProcessMessages;
    ScaleVector(Shade,Cover);
    Application.ProcessMessages;
    AddVector(Shade,VectorScale(Cover,fAmbientLight));
    Application.ProcessMessages;
    if Shade[0]>1 then Shade[0]:=1;
    if Shade[1]>1 then Shade[1]:=1;
    if Shade[2]>1 then Shade[2]:=1;
    //if x=Side-1 then begin Shade[0]:=1; Shade[1]:=0; Shade[2]:=0; end;
    MakeRGBTriple(Shade,Result);
  end;

begin
  nbTiles:=fSize div FTerrainRenderer.TileSize;
  SetLength(Bmp,nbTiles,nbTiles);
  Side:=FTerrainRenderer.TileSize*TextureScale;
  fTask:='Texture creation';
  fTaskProgress:=0;

  {Draw bitmap}
  try
    for yy:=0 to (nbTiles)-1 do begin
      fTaskProgress:=Round(yy/nbTiles*100);
      Application.ProcessMessages;
      y0:=yy*Side;
      for xx:=0 to (nbTiles)-1 do begin
        x0:=xx*Side;
        Bmp[xx,yy]:=TBitmap.Create;
        with Bmp[xx,yy] do begin
          PixelFormat:=pf24bit;
          Width:=Side;
          Height:=Side;
          for y:=0 to Side-1 do begin
            Line:=ScanLine[y];
            for x:=0 to Side-1 do begin
              Line[x]:=ComputePixel(x,y);
            end;//for x
          end;//for y
        end;//with
      end;//for xx
    end;//for yy

    {Smoothes tile seams}
    for yy:=0 to nbTiles-2 do begin
      for xx:=0 to nbTiles-2 do begin
        for x:=0 to Side-1 do begin
          meancol:=MeanColor(Bmp[xx,yy].Canvas.Pixels[Side-1,x],
                             Bmp[xx+1,yy].Canvas.Pixels[0,x]);

          Bmp[xx,yy].Canvas.Pixels[Side-1,x]:=meancol;
          Bmp[xx+1,yy].Canvas.Pixels[0,x]:=meancol;
          meancol:=MeanColor(Bmp[xx,yy].Canvas.Pixels[x,Side-1],
                             Bmp[xx,yy+1].Canvas.Pixels[x,0]);

          Bmp[xx,yy].Canvas.Pixels[x,Side-1]:=meancol;
          Bmp[xx,yy+1].Canvas.Pixels[x,0]:=meancol;
        end;//for x
      end;//for xx
    end;//for yy

    {Upload into material library}
    for yy:=0 to nbTiles-1 do begin
      for xx:=0 to nbTiles-1 do begin
        with FTerrainRenderer.MaterialLibrary
        .AddTextureMaterial(Format('%s%d%d',[Self.Name,xx,yy]),Bmp[xx,yy]) do begin
          //Material.Texture.MinFilter:=miNearest;
          Material.Texture.TextureWrap:=twNone;
          Material.MaterialOptions:=[moNoLighting]; // Needed for correct look when lighting is enabled
        end;//with
        //Bmp[xx,yy].SaveToFile(Format('%s%d%d.bmp',[Self.Name,xx,yy]));
      end;//for xx
    end;//for yy
  finally
    for yy:=0 to nbTiles-2 do begin
      for xx:=0 to nbTiles-2 do begin
        Bmp[xx,yy].Free;
      end;//for xx
    end;//for yy
    Bmp:=nil;
  end;//finally
end; //*)

(*procedure tGLCustomRandomHDS.BuildTexture2;
var
  Bmp	:tBitmap;
  Mat	:TGLLibMaterial;
  x,y	:integer;
  i,j	:integer;
  Shade	:tColorVector;
  Cover	:tColorVector;
  z		:double;
begin
  if not fTextureCreated then CreateTexture;
  Mat:=FTerrainRenderer.MaterialLibrary.LibMaterialByName(MaterialName);
  Bmp:=TBitmap.Create;
  fTask:='Texture creation';
  fTaskProgress:=0;

  {Draw bitmap}
  try
    with Bmp do begin
      PixelFormat:=pf24bit;
      Width:=fSize*TextureScale;
      Height:=fSize*TextureScale;
      with Canvas do begin
        for y:=0 to fSize*TextureScale-1 do begin
          fTaskProgress:=Round(y/(fSize*TextureScale)*100);
          Application.ProcessMessages;
          for x:=0 to fSize*TextureScale-1 do begin
            i:=x div TextureScale;
            j:=y div TextureScale;
            z:=Interpolate(x/TextureScale,y/TextureScale);
            Cover:=OnDrawTexture(Self,x,y,z,fNormal[i,j]);
            Shade:=VectorScale(fLightColor.Color,fLightMap[x,y]);
            ScaleVector(Shade,Cover);
            AddVector(Shade,VectorScale(Cover,fAmbientLight));

            if Shade[0]>1 then Shade[0]:=1;
            if Shade[1]>1 then Shade[1]:=1;
            if Shade[2]>1 then Shade[2]:=1;
            Pixels[x,y]:=ConvertColorVector(Shade);
          end;//for x
        end;//for y
      end;//with
    end;//with
    //Bmp.SaveToFile('test.bmp');

    {Use it as texture}
    with Mat.Material.Texture do begin
      Image.Assign(Bmp);
      Image.NotifyChange(Self);
      Enabled:=true;
      //MagFilter:=maNearest;
      //MinFilter:=miNearest;
    end;//with }
    Mat.NotifyUsersOfTexMapChange;
  finally
    Bmp.Free;
  end;//finally
end;//  *)

procedure tGLCustomRandomHDS.ClearHeightField;
begin
  PrimerNull(fHeight);
end;

procedure tGLCustomRandomHDS.ClearLightMap;
var
  x,y	:integer;
begin
  SetLength(fLightMap,(fSize+1)*TextureScale,(fSize+1)*TextureScale);
  for y:=0 to High(fLightMap) do begin
    for x:=0 to High(fLightMap) do begin
      fLightMap[x,y]:=1;
    end;//for
  end;//for
end;

procedure tGLCustomRandomHDS.ConstrainCoordinates(var x, y: integer);
begin
  fIntegerConstrain(x,y);
end;

procedure tGLCustomRandomHDS.ConstrainCoordinates(var x, y: single);
begin
  fSingleConstrain(x,y);
end;

constructor tGLCustomRandomHDS.Create(AOwner:tComponent);
begin
  inherited;
  FLandCover:=True;
  FOnDrawTexture:=OnDrawTextureDefault;
end;

constructor tGLCustomRandomHDS.CreateAsSlave(AOwner: TComponent);
begin
  fSlave:=True;
  Create(AOwner);
end;

procedure tGLCustomRandomHDS.CreateTexture;
begin
  if FTerrainRenderer=nil then begin
    raise EInvalidOperation.Create('No TerrainRenderer linked to this component.');
    exit;
  end;//if
  if FTerrainRenderer.MaterialLibrary=nil then begin
    raise EInvalidOperation.Create('No material library assigned to the TerrainRenderer');
    exit;
  end;//if
  with FTerrainRenderer.MaterialLibrary do begin
    (Materials.Add).Name:=fTerrainRenderer.Name+'Light';
  end;//with
  MaterialName:=fTerrainRenderer.Name+'Light';
  fTextureCreated:=True;
end;

procedure tGLCustomRandomHDS.CyclicClamp(var x, y: single);
var
  ix,iy	:integer;
  sx,sy	:single;
begin
  ix:=Trunc(x);  sx:=Frac(x);
  iy:=Trunc(y);  sy:=Frac(y);
  x:=(fSize+ix)mod fSize+sx;
  y:=(fSize+iy)mod fSize+sy;
end;

procedure tGLCustomRandomHDS.CyclicClamp(var x, y: integer);
begin
  x:=(fSize+x)mod fSize;
  y:=(fSize+y)mod fSize;
end;

destructor tGLCustomRandomHDS.Destroy;
var
  x,y	:integer;
  Mat	:TGLLibMaterial;
begin
  with LandTileInfo do State:=hdsNone;
  fHeight:=nil;
  fLightMap:=nil;
  fNormal:=nil;
  try
    for y:=0 to (fSize div FTerrainRenderer.TileSize)-1 do begin
      for x:=0 to (fSize div FTerrainRenderer.TileSize)-1 do begin
        Mat:=FTerrainRenderer.MaterialLibrary.LibMaterialByName(Format('%s%d%d',[Self.Name,x,y]));
        if Mat<>nil then Mat.Material.DestroyHandles;
      end;//for x
    end;//for y
  except end;
  if (fSlave)and(Owner<>nil) then with LandTileInfo
  do tGLTiledRndLandscape(Owner).MarkDirty(x*fSize,z*fSize,(x+1)*fSize-1,(z+1)*fSize-1);
  inherited;
end;

procedure tGLCustomRandomHDS.DoCyclicBoundaries;
var
  i	:integer;
begin
  for i:=0 to fSize do begin
    fHeight[i,fSize]:=fHeight[i,0];
    fHeight[fSize,i]:=fHeight[0,i];
  end;//for
end;

procedure tGLCustomRandomHDS.DoErosionByFraction;
begin

end;

procedure tGLCustomRandomHDS.DoErosionByLife;
var
  x,y,i		:integer;
  z,z1		:single;
begin
  {Smoothing by a 3-by-3 mean filter}
  fTask:='Erosion by life';
  fTaskProgress:=0;
  for y:=0 to fSize do begin
    fTaskProgress:=Round(y/(fSize)*100);
    for x:=0 to fSize do begin
      Application.ProcessMessages;
      z:=fHeight[x,y]*fErosionByLife.Robustness;
      z1:=fErosionByLife.Robustness;
      for i:=0 to 7 do begin
        z:=z+Height[x+NeighX[i],y+NeighY[i]]*NeighW[i];
        z1:=z1+NeighW[i];
      end;//for i
      fHeight[x,y]:=z/z1;
    end;//for x
  end;//for y
end;

procedure tGLCustomRandomHDS.DoErosionByRain;
{Drop some rain on every cell of the landscape and let it run downward, taking soil
 on its way. When it arrives into a pool, let it deposit all that has been eroded.}
const
  Ks=0.001; // Soil solubility
var
  j			:integer;
  x0,y0		:integer;
  x,y		:integer;
  x1,y1		:integer;
  minx,miny	:integer;
  z,z1		:single;
  minz		:single;
  dz,mindz	:single;
  Charge		:double;
  From,Next	:integer;
begin
  fTask:='Rain erosion simulation';
  fTaskProgress:=0;
  minx:=0; miny:=0; minz:=0;Next:=0;

  {Rain}
  for y0:=0 to fSize do begin
    fTaskProgress:=Round(y0/(fSize)*100);
    for x0:=0 to fSize do begin
      Application.ProcessMessages;
      x:=x0;
      y:=y0;
      z:=StandardisedHeight(x,y);
      Charge:=0;
      From:=-1;
      while (fHeight[x,y]>fSeaLevel)  // Not in the sea
      do begin
        mindz:=MaxInt;
        for j:=0 to 7 do begin // Look for the largest slope
          if j=From then continue; // Never go backward
          x1:=(fSize+x+NeighX[j]) mod fSize; // Cyclic landscape
          y1:=(fSize+y+NeighY[j]) mod fSize;
          z1:=StandardisedHeight(x1,y1);
          dz:=(z1-z)*NeighW[j];
          if dz<mindz then begin
            minx:=x1;
            miny:=y1;
            minz:=z1;
            mindz:=dz;
            Next:=j;
          end;//if
        end;//for j
        if (StandardisedHeight(minx,miny)<=SeaLevel) then break; // In the sea or out of map
        if minz<z then begin
          fHeight[x,y]:=fHeight[x,y]-FErosionByRain.ErosionRate*Ks*fRangeHeight; // Erosion
          x:=minx;
          y:=miny;
          z:=minz;
          From:=(Next+4)mod 8; // Opposite direction
          Charge:=Charge+1 ;
          end //if
        else begin // Fallen into a pool? Deposit the charge
          fHeight[x,y]:=fHeight[x,y]+MinFloat(minz-z,FErosionByRain.DepositRate*Ks*fRangeHeight
                                     *Charge);
          Break; // Go to next rain drop
        end;//else
      end;//while
    end;//for x0
  end;//for y0
end; // *)

(*procedure tGLBaseRandomHDS.DoErosionByRain(const Intensity: single);
const
  NeighX	:array[0..7] of integer=(-1, 0, 1, 1, 1, 0,-1,-1);
  NeighY	:array[0..7] of integer=(-1,-1,-1, 0, 1, 1, 1, 0);
  NeighW   :array[0..7] of single=(1/1.4142,1,1/1.4142,1,1/1.4142,1,1/1.4142,1);
type
  tFlow=record
    NextX,NextY	:integer;
    Slope		:single;
    Erosion		:integer;
  end;
var
  Flow		:array of array of tFlow;
  i,j,jj,swap	:integer;
  x0,y0		:integer;
  x,y		:integer;
  x1,y1		:integer;
  minx,miny	:integer;
  z,z1,minz	:single;
  Charge		:integer;
  N			:integer;
  From,Next	:integer;
  Sig		:integer;
  c			:double;
  OldSlope	:single;
  dz,mindz	:single;
begin
  c:=1/VSF/sqrt(sqr(Scale.X)+sqr(Scale.Z));
  {Water flow map computation}
  SetLength(Flow,fSize+1,fSize+1);
  for y:=0 to fSize do begin
    for x:=0 to fSize do begin
      mindz:=MaxInt;
      Sig:=Sign(random*2-1);
      z:=fHeight[x,y];
      for jj:=0 to 7 do begin // Look for the largest slope
        j:=(8+Sig*jj) mod 8;
        x1:=x+NeighX[j];
        y1:=y+NeighY[j];
        try z1:=Height[x1,y1];
          dz:=(z1-z)*NeighW[j];
          if dz+random*0.03*fRangeHeight<mindz then begin
            minx:=x1;
            miny:=y1;
            minz:=z1;
            mindz:=dz;
            Next:=j;
          end;//if
        except // Out of the map? Then go to next rain drop
          Flow[x,y].NextX:=-99;
          Break;
        end;
        with Flow[x,y] do begin
          Slope:=ArcTan((minz-z)*c);
          if Slope>0 then NextX:=-99
          else begin
            NextX:=minx;
            NextY:=miny;
            Erosion:=0;
          end;//if
        end;//with
      end;//for j
    end;//for
  end;//for
  From:=0;

  {Rain}
  for y0:=0 to fSize do begin
    for x0:=0 to fSize do begin
      x:=x0;
      y:=y0;
      OldSlope:=0;
      while (x<>-99)and(fCover[x,y]>0) do begin // Not in the sea
        with Flow[x,y] do begin
          if (Slope*2<OldSlope) then begin
            Dec(Erosion);
            x:=NextX;
            y:=NextY;
            OldSlope:=Slope;
            end//if
          else begin
            //Inc(Erosion);
            Break;
          end;//else
        end;//with
      end;//while
    end;//for x0
  end;//for y0

  {Apply erosion}
  for y:=0 to fSize do begin
    for x:=0 to fSize do begin
      //fHeight[x,y]:=fHeight[x,y]+Flow[x,y].Erosion*0.002*Intensity*fRangeHeight;
      fHeight[x,y]:=(Flow[x,y].Erosion)*100+50;
    end;//for
  end;//for

  Flow:=nil;
end; // *)

(*procedure tGLBaseRandomHDS.DoErosionByRain(const Intensity: single);
const
  NeighX:array[0..7] of integer=(-1, 0, 1, 1, 1, 0,-1,-1);
  NeighY:array[0..7] of integer=(-1,-1,-1, 0, 1, 1, 1, 0);
var
  Erosion	:array of array of single;
  Flow		:array[0..7] of single;
  FlowSum	:single;
  j			:integer;
  x,y		:integer;
  x1,y1		:integer;
  z,z1		:single;
  c			:single;
begin
  c:=1/VSF; // Vertical scale factor
  SetLength(Erosion,fSize+2,fSize+2);
  for y:=0 to fSize+1 do for x:=0 to fSize+1 do Erosion[x,y]:=0;

  {Erosion computation}
  for y:=0 to fSize+1 do begin
    for x:=0 to fSize+1 do begin
      z:=fHeight[x,y];
      FlowSum:=0;
      for j:=0 to 7 do begin // Flow to adjacent cells
        x1:=x+NeighX[j];
        y1:=y+NeighY[j];
        try
          z1:=Height[x1,y1]+random*0;
          if z1<z then begin
            Flow[j]:=ArcTan((z-z1)*c);
            FlowSum:=FlowSum+Flow[j];
            end//if
          else Flow[j]:=0;
        except
          Flow[j]:=0;
        end;//except
      end;//for j

      if FlowSum>0 then begin // Erosion and deposition
        Erosion[x,y]:=Erosion[x,y]-1; // Erosion
        for j:=0 to 7 do begin
          if Flow[j]>1e-3 then begin
            x1:=x+NeighX[j];
            y1:=y+NeighY[j];
            Erosion[x1,y1]:=Erosion[x1,y1]+Flow[j]/FlowSum; // Partial deposition
          end;//if
        end;//for
      end;//if

    end;//for x
  end;//for y

  {Apply erosion to each cell}
  for y:=0 to fSize do begin
    for x:=0 to fSize do begin
      fHeight[x,y]:=fHeight[x,y]+Erosion[x,y]*0.005*Intensity*fRangeHeight;
      //fHeight[x,y]:=(Erosion[x,y])*100+50;
    end;//for
  end;//for

  Erosion:=nil;
end; // *)

(*procedure tGLBaseRandomHDS.DoErosionByRain(const Intensity: single);
const
  NeighX:array[0..7] of integer=(-1, 0, 1, 1, 1, 0,-1,-1);
  NeighY:array[0..7] of integer=(-1,-1,-1, 0, 1, 1, 1, 0);
var
  Erosion	:array of array of single;
  i,j,jj		:integer;
  x,y		:integer;
  x1,y1		:integer;
  x2,y2		:integer;
  z,z1,z2,dz	:single;
begin
  SetLength(Erosion,fSize+2,fSize+2);

  for i:=1 to 1 do begin
    //for y:=0 to fSize+1 do for x:=0 to fSize+1 do Erosion[x,y]:=0;

    {Erosion computation}
    for y:=5 to fSize-4 do begin
      for x:=5 to fSize-4 do begin
        z:=fHeight[x,y];
        dz:=1;
        for jj:=1 to 2 do begin // Flow to adjacent cells
          j:=jj*2;
          x1:=x+NeighX[j]*5;
          y1:=y+NeighY[j]*5;
          x2:=x+NeighX[j+4]*5;
          y2:=y+NeighY[j+4]*5;
          try
            z1:=Height[x1,y1]+random*0;
            z2:=Height[x2,y2]+random*0;
            dz:=dz*Sign(z-(z1+z2)/2);
          except
          end;//except
        end;//for j
        Erosion[x,y]:=dz;
      end;//for x
    end;//for y

    {Apply erosion to each cell}
    for y:=0 to fSize do begin
      for x:=0 to fSize do begin
        fHeight[x,y]:=fHeight[x,y]+Erosion[x,y]*100*Intensity;
        //fHeight[x,y]:=(Erosion[x,y])*1+50;
      end;//for
    end;//for

  end;//for i

  Erosion:=nil;
end; // *)

procedure tGLCustomRandomHDS.DoErosionBySea;
var
  i,j	:integer;
begin
  for i:=0 to fSize do begin
    for j:=0 to fSize do begin
      Application.ProcessMessages;
      if abs(fHeight[i,j]-fSeaLevel)<FErosionBySea.BeachHeight*VSF then begin
        fHeight[i,j]:=fSeaLevel+(fHeight[i,j]-fSeaLevel)*0.3;
      end;//if
    end;//for
  end;//for
end;

procedure tGLCustomRandomHDS.DoSea;
var
  i,j	:integer;
begin
  for i:=0 to fSize do begin
    for j:=0 to fSize do begin
      //if fHeight[i,j]<Lvl then fHeight[i,j]:=Lvl-random*wave;
      if fHeight[i,j]<fSeaLevel-fSeaTransparency
      then fHeight[i,j]:=fSeaLevel-1//Lvl-c-random*wave
      else if fHeight[i,j]<fSeaLevel
      then fHeight[i,j]:=fSeaLevel-(fSeaLevel-fHeight[i,j])/fSeaTransparency;
    end;//for
  end;//for
end;

procedure tGLCustomRandomHDS.DoSteps;
var
  i,j	:integer;
  Stp	:single;
begin
  Stp:=(fMaxHeight-FSeaLevel)/FSteps.Count; // Step height
  for i:=0 to fSize do begin
    for j:=0 to fSize do begin
      fHeight[i,j]:=Round(fHeight[i,j]/Stp)*Stp;
    end;//for
  end;//for
end;

function tGLCustomRandomHDS.GetHeight(x, y: integer): single;
begin
  fIntegerConstrain(x,y);
  Result:=fHeight[x,y];
end;

procedure tGLCustomRandomHDS.GetTerrainBounds(var l, t, r, b: Single);
begin
  l:=0;
  b:=0;
  t:=fSize;
  r:=fSize;
end;

function tGLCustomRandomHDS.Interpolate(x, y: single): single;
{Copied from GLHeightData.InterpolatedHeight}
var
  ix, iy : Integer;
  h1, h2, h3 : Single;
begin
  ix:=Trunc(x);  x:=Frac(x);
  iy:=Trunc(y);  y:=Frac(y);
  if x > y then begin
    // top-right triangle
    h1:=Height[ix+1,  iy];
    h2:=Height[ix,    iy];
    h3:=Height[ix+1,  iy+1];
    Result:=h1+(h2-h1)*(1-x)+(h3-h1)*y;
  end else begin
    // bottom-left triangle
    h1:=Height[ix,    iy+1];
    h2:=Height[ix+1,  iy+1];
    h3:=Height[ix,    iy];
    Result:=h1+(h2-h1)*(x)+(h3-h1)*(1-y);
  end;
end;

function tGLCustomRandomHDS.PointInMap(const x, y: single): boolean;
begin
  Result:=(x>=0)and(x<=fSize)and(y>=0)and(y<=fSize);
end;

function tGLCustomRandomHDS.Normal(Position:tVector): tVector;
var
  x,y	:integer;
begin
  if (fNormal<>nil) then begin
    Result:=FTerrainRenderer.AbsoluteToLocal(Position);
    x:=Round(Result[0]);
    y:=Round(Result[1]);
    fIntegerConstrain(x,y);
    Result:=fNormal[x,y];
    end//if
  else raise EAccessViolation.Create('No normal array computed.');
end;

function tGLCustomRandomHDS.PointInMap(const x, y: integer): boolean;
begin
  Result:=(x>=0)and(x<=fSize)and(y>=0)and(y<=fSize);
end;

function tGLCustomRandomHDS.Scale: TGLCoordinates;
begin
  try
    Result:=FTerrainRenderer.Scale;
  except
    raise EAccessViolation.Create('No TerrainRenderer defined');
  end;
end;

procedure tGLCustomRandomHDS.SetCyclic(const Value: boolean);
begin
  FCyclic := Value;
  if fCyclic then begin
    fIntegerConstrain:=CyclicClamp;
    fSingleConstrain:=CyclicClamp;
    if FTerrainRenderer<>nil then FTerrainRenderer.OnGetTerrainBounds:=nil;
    end
  else begin
    fIntegerConstrain:=BoundaryClamp;
    fSingleConstrain:=BoundaryClamp;
    if FTerrainRenderer<>nil then FTerrainRenderer.OnGetTerrainBounds:=GetTerrainBounds;
  end;//else
end;

procedure tGLCustomRandomHDS.SetHeight(x, y: integer;
  const Value: single);
begin
  fIntegerConstrain(x,y);
  fHeight[x,y]:=Value;
end;

procedure tGLCustomRandomHDS.SetSize(const aSize: integer);
var
  Tile	:integer;
begin
  fSize:=aSize;
  if fSize>32 then Tile:=32
  else Tile:=Round(IntPower(2,Trunc(ln(fSize-1)/ln(2))));
  SetLength(fHeight,fSize+1,fSize+1);
  SetLength(fNormal,fSize+1,fSize+1);
  MaxPoolSize:=sqr(fSize)*SizeOf(smallint);
  if FTerrainRenderer<>nil then begin
    FTerrainRenderer.TileSize:=Tile;
    fTerrainRenderer.TilesPerTexture:=fSize div FTerrainRenderer.TileSize;
  end;//if
end;

procedure tGLCustomRandomHDS.SetTerrainRenderer(
  const Value: TGLTerrainRenderer);
begin
  FTerrainRenderer := Value;
  if not fSlave then begin
    FTerrainRenderer.OnGetTerrainBounds:=GetTerrainBounds;
    FTerrainRenderer.HeightDataSource:=Self;
  end;//if
end;

function tGLCustomRandomHDS.StandardisedHeight(const x, y: integer): single;
begin
  Result:=(Height[x,y]-fMinHeight)/fRangeHeight*1000;
end;

procedure tGLCustomRandomHDS.StartPreparingData(heightData: THeightData);
var
  x,y,x0,y0	:Integer;
  rasterLine	:GLHeightData.PSmallIntArray;
  oldType 	:THeightDataType;
begin
  with heightData do begin
    DataState:=hdsPreparing;
    oldType:=DataType;
    Allocate(hdtSmallInt);

    if XLeft>=0 then x0:=XLeft mod (fSize) else x0:=(fSize+(XLeft mod (fSize)))mod (fSize);
    if YTop>=0 then y0:=YTop mod (fSize) else y0:=(fSize+(YTop mod (fSize)))mod (fSize);
    if fLandCover then begin
      MaterialName:=Format('%s%d%d',[Self.Name,x0 div (HeightData.Size-1),y0 div (HeightData.Size-1)]);
      TextureCoordinatesMode:=tcmLocal;
      TextureCoordinatesScale:=TexPointMake((Self.fSize)/(HeightData.Size-1),(Self.fSize)/(HeightData.Size-1));
      end//if
    else begin
      MaterialName:=Self.FMaterialName;
      TextureCoordinatesMode:=tcmLocal;
      TextureCoordinatesScale:=TexPointMake(FTextureScale,FTextureScale);
    end;//else
    for y:=y0 to y0+heightData.Size-1 do begin
      rasterLine:=smallintRaster[y-y0];
      for x:=x0 to x0+heightData.Size-1 do begin
        rasterLine[x-x0]:=Round(fHeight[x,y]);
      end;//for
    end;//for
    HeightMin:=MinHeight;
    HeightMax:=MaxHeight;

    DataState:=hdsReady;
    if oldType<>hdtSmallint then DataType:=oldType;
  end;//with
  //inherited;
end; //*)

function tGLCustomRandomHDS.XMoveBoundary: single;
begin
  Result:=fSize*Scale.X*0.95;
end;

function tGLCustomRandomHDS.ZMoveBoundary: single;
begin
  Result:=fSize*Scale.Y*0.95;
end;

procedure tGLCustomRandomHDS.SetKeepNormals(const Value: boolean);
begin
  FKeepNormals := Value;
end;

{ tGLFractalHDS }

procedure tGLFractalHDS.BuildHeightField(const aDepth, aSeed,aAmplitude: integer);
begin
  fDepth:=aDepth;
  fSeed:=aSeed;
  fAmplitude:=aAmplitude;
  BuildHeightField;
end;

procedure tGLFractalHDS.BuildHeightField;
begin
  FractalMiddlePointHDS(fDepth,fSeed,FAmplitude,fRoughness,fCyclic,fHeight,fMinHeight,fMaxHeight);
  fRangeHeight:=fMaxHeight-fMinHeight;
  Scale.X:=1;
  Scale.Y:=1;
  Scale.Z:=fSize/VSF;
end;

constructor tGLFractalHDS.Create(AOwner:tComponent);
begin
  inherited;
  Depth:=4;
  fSea:=True;
  Amplitude:=50;
  fRoughness:=0.4;
end;

procedure tGLFractalHDS.SetAmplitude(const Value: integer);
begin
  fAmplitude := Value;
  fMinHeight:=-fAmplitude/2*VSF;
  fMaxHeight:=-fMinHeight;
  fRangeHeight:=fAmplitude*VSF;
end;

procedure tGLFractalHDS.SetDepth(const Value: integer);
begin
  fDepth := Value;
  SetSize(Round(IntPower(2,fDepth)));
end;

procedure tGLFractalHDS.SetRoughness(const Value: single);
begin
  FRoughness := Value;
end;

{ tGLRandomLandscape }

procedure tGLTiledRndLandscape.ApplyLighting(var aLandTile: tLandTile);
begin
  with aLandTile do begin
    Lighting:=Self.fLighting;
    LightColor:=Self.fLightColor;
    LightDirection:=Self.fLightDirection;
    LightSmoothing:=Self.fLightSmoothing;
    Shadows:=Self.Shadows;
  end;//with
end;

procedure tGLTiledRndLandscape.ApplyTexture(var aLandTile: tLandTile);
begin
  with aLandTile do begin
    LandCover:=Self.LandCover;
    MaterialName:=Self.FMaterialName;
    TextureScale:=Self.fTextureScale;
    if Assigned(Self.OnDrawTexture) then FOnDrawTexture:=Self.OnDrawTexture;
  end;//with
end;

procedure tGLTiledRndLandscape.ApplyTopography(var aLandTile: tLandTile);
begin
  with aLandTile do begin
    ErosionByFraction:=Self.fErosionByFraction;
    ErosionByLife:=Self.fErosionByLife;
    ErosionByRain:=Self.FErosionByRain;
    ErosionBySea:=Self.FErosionBySea;
    fSea:=Self.fSea;
    fSeaLevel:=Self.FSeaLevel;
    fSeaTransparency:=Self.FSeaTransparency;
  end;//with
end;

procedure tGLTiledRndLandscape.BoundaryClamp(var x, z: single);
begin
  ClampValue(x,0,fExtentX*fLandTileSize);
  ClampValue(z,0,fExtentZ*fLandTileSize);
end;

procedure tGLTiledRndLandscape.BoundaryClamp(var x, z: integer);
begin
  if x<0 then x:=0 else if x>fExtentX*fLandTileSize
  then x:=fExtentX*fLandTileSize;
  if z<0 then z:=0 else if z>ExtentZ*fLandTileSize
  then z:=fExtentZ*fLandTileSize;
end;

procedure tGLTiledRndLandscape.CameraPosition(var TileX, TileZ: integer);
begin
  FindLandTile(-Camera.Position.X,Camera.Position.Z,TileX,TileZ);
end;

procedure tGLTiledRndLandscape.CleanUp;
var
  i	:integer;
begin
  for i:=fLandTiles.Count-1 downto 0 do begin
    if tLandTile(fLandTiles.Items[i]).LandTileInfo.State=hdsNone then begin
      fLandTiles.Delete(i); // Free the landtile and remove it from the list
      //FTerrainRenderer.MaterialLibrary.Materials.DeleteUnusedMaterials;
    end;//if
  end;//for
end;

procedure tGLTiledRndLandscape.ComputeLandTile(const aX,aZ: integer;
                                               var NewLandTile:tLandTile);
var
  sx,sz	:string;
begin
  fLandTileComputing:=True;
  with NewLandTile.LandTileInfo do begin
    x:=aX;
    z:=aZ;
    State:=hdsPreparing;
  end;//with
  with NewLandTile do begin
    Cyclic:=False;
    TerrainRenderer:=Self.fTerrainRenderer;
    if ax>=0 then sx:='p' else sx:='n';
    if az>=0 then sz:='p' else sz:='n';
    Seed:=LandTileSeed(ax,az);
    Name:=Format('Land_%s%d%s%d_',[sx,abs(ax),sz,abs(az)]); // Generate a unique name
  end;//with
  fComputedLandTile:=NewLandTile;
  OnCreateLandTile(ax,az,NewLandTile.Seed,NewLandTile);
  with NewLandTile.LandTileInfo do State:=hdsReady;
  MarkDirty(ax*fLandTileSize,az*fLandTileSize,(ax+1)*fLandTileSize-1,(az+1)*fLandTileSize-1);
  fComputedLandTile:=nil;
  fLandTileComputing:=False;
  fLandTiles.Add(NewLandTile);
  Application.ProcessMessages;
end;

procedure tGLTiledRndLandscape.ConstrainCoordinates(var x, z: single);
begin
  fSingleConstrain(x,z);
end;

procedure tGLTiledRndLandscape.ConstrainCoordinates(var x, z: integer);
begin
  fIntegerConstrain(x,z);
end;

constructor tGLTiledRndLandscape.Create(AOwner: TComponent);
begin
  inherited;
  fLandTiles:=TComponentList.Create;
  IsDefaultTile:=fDefaultIsDefaultTile;
  OnCreateDefaultTile:=fDefaultOnCreateDefaultTile;
  FExtentX:=10000;
  FExtentZ:=10000;
  GenerationRadius:=2;
  FLandTileDensity:=1;
  FLandCover:=True;
end;

procedure tGLTiledRndLandscape.CyclicClamp(var x, z: integer);
begin
  Exit;
  x:=(x+ExtentX) mod ExtentX;
  z:=(z+ExtentZ) mod ExtentZ;
end;

procedure tGLTiledRndLandscape.CyclicClamp(var x, z: single);
var
  ix,iz	:integer;
  sx,sz	:single;
begin
  Exit;
  ix:=Trunc(ExtentX+x);  sx:=Frac(x);
  iz:=Trunc(ExtentZ+z);  sz:=Frac(z);
  x:=(ExtentX*fLandTileSize+ix)mod ExtentX*fLandTileSize +sx;
  z:=(ExtentZ*fLandTileSize+iz)mod ExtentZ*fLandTileSize +sz;
end;

destructor tGLTiledRndLandscape.Destroy;
begin
  fLandTiles.Free;
  inherited;
end;

function tGLTiledRndLandscape.fDefaultIsDefaultTile(x, z: integer): boolean;
begin
  InitializeRandomGenerator(LandTileSeed(x,z));
  Result:=(random>=fLandTileDensity);
end;

procedure tGLTiledRndLandscape.fDefaultOnCreateDefaultTile(HeightData: THeightData);
begin
  HeightData.DataState:=hdsNone;
//  raise EAccessViolation.Create('No DefaultStartPreparingDefaultTile procedure supplied.');
end;

procedure tGLTiledRndLandscape.FindLandTile(const x, z: single; var TileX,
  TileZ: integer);
begin
  TileX:=Floor(x/fLandTileSize);
  TileZ:=Floor(z/fLandTileSize);
  fIntegerConstrain(TileX,TileZ);
end;

function tGLTiledRndLandscape.fSortLandscapes(Item1,Item2: Pointer): Integer;
{Sort landtiles from the closest to the farthest}
var
  x,z	:integer;
  d1,d2	:single;
begin
  CameraPosition(x,z);
  d1:=sqr(x-tLandTile(Item1^).LandTileInfo.x)+sqr(z-tLandTile(Item1^).LandTileInfo.z);
  d2:=sqr(x-tLandTile(Item2^).LandTileInfo.x)+sqr(z-tLandTile(Item2^).LandTileInfo.z);
  Result:=Round(d1-d2);
end;

function tGLTiledRndLandscape.GetTask: string;
begin
  if fComputedLandTile<>nil
  then Result:=fComputedLandTile.Task
  else Result:='Idle';
end;

function tGLTiledRndLandscape.GetTaskProgress: integer;
begin
  if fComputedLandTile<>nil
  then Result:=fComputedLandTile.TaskProgress
  else Result:=0;
end;

procedure tGLTiledRndLandscape.GetTerrainBounds(var l, t, r, b: Single);
begin
  l:=0;
  b:=0;
  t:=ExtentZ*LandTileSize;
  r:=ExtentX*LandTileSize;
end;

procedure tGLTiledRndLandscape.Initialize(const ax, az: single);
var
  cx,cz		:integer;
  NewLandTile	:tLandTile;
  x,z,dx,dz	:integer;
begin
  fOldCamX:=-99999;
  fOldCamZ:=-99999;
  with Camera.Position do begin
    X:=ax;
    Z:=aZ;
  end;//with
  CameraPosition(cx,cz);
  ComputeLandTile(cx,cz,NewLandTile);
  TerrainRenderer.Scale:=NewLandTile.Scale;
  with Camera.Position do begin
    X:=X*NewLandTile.Scale.X;
    Z:=Z*NewLandTile.Scale.Z;
  end;//with

  for z:=0 to FGenerationRadius+1 do begin
    for x:=1 to FGenerationRadius+1 do begin
      if Trunc(sqrt(sqr(x)+sqr(z)))<=FGenerationRadius then begin
        dx:=x;  	dz:=z;
        if not IsDefaultTile(cx+dx,cz+dz)
        then ComputeLandTile(cx+dx,cz+dz,NewLandTile);
        dx:=-z; 	dz:=x;
        if not IsDefaultTile(cx+dx,cz+dz)
        then ComputeLandTile(cx+dx,cz+dz,NewLandTile);
        dx:=-x; 	dz:=-z;
        if not IsDefaultTile(cx+dx,cz+dz)
        then ComputeLandTile(cx+dx,cz+dz,NewLandTile);
        dx:=z; 	dz:=-x;
        if not IsDefaultTile(cx+dx,cz+dz)
        then ComputeLandTile(cx+dx,cz+dz,NewLandTile);
      end;//if
    end;//for
  end;//for
end;

function tGLTiledRndLandscape.LandTileSeed(x, z: integer): integer;
{Generates a unique seed from the tile coordinates}
begin
  Result:=fBaseSeed + z*ExtentX+x;
end;

function tGLTiledRndLandscape.LandtileCount: integer;
begin
  Result:=fLandTiles.Count;
end;

{procedure tGLTiledRndLandscape.PrepareLandTileData(HeightData: tHeightData;
  LandTile: tLandTile);
var
  x,y,x0,y0	:integer;
  rasterLine	:GLHeightData.PSingleArray;
  oldType 	:THeightDataType;
begin
  with HeightData do begin
    DataState:=hdsPreparing;
    oldType:=DataType;
    Allocate(hdtSingle);

    if XLeft>=0 then x0:=XLeft mod (fLandTileSize) else x0:=(fLandTileSize+(XLeft mod (fLandTileSize)))mod (fLandTileSize);
    if YTop>=0 then y0:=YTop mod (fLandTileSize) else y0:=(fLandTileSize+(YTop mod (fLandTileSize)))mod (fLandTileSize);

    MaterialName:=Format('%s%d%d',[LandTile.Name,x0 div fTerrainRenderer.TileSize,
                                                 y0 div fTerrainRenderer.TileSize]);
    TextureCoordinatesMode:=tcmLocal;
    TextureCoordinatesScale:=TexPointMake((fLandTileSize)/(Size),
                                          (fLandTileSize)/(Size));
    for y:=y0 to y0+heightData.Size-1 do begin
      rasterLine:=singleRaster[y-y0];
      for x:=x0 to x0+heightData.Size-1 do begin
        rasterLine[x-x0]:=LandTile.fHeight[x,y];
      end;//for
    end;//for

    DataState:=hdsReady;
    if oldType<>hdtSingle then DataType:=oldType;
  end;//with
end; }

procedure tGLTiledRndLandscape.SetCamera(const Value: TGLCamera);
begin
  FCamera := Value;
end;

procedure tGLTiledRndLandscape.SetCyclic(const Value: boolean);
begin
  FCyclic := Value;
  if fCyclic then begin
    fIntegerConstrain:=CyclicClamp;
    fSingleConstrain:=CyclicClamp;
    if FTerrainRenderer<>nil then FTerrainRenderer.OnGetTerrainBounds:=nil;
    end
  else begin
    fIntegerConstrain:=BoundaryClamp;
    fSingleConstrain:=BoundaryClamp;
    if FTerrainRenderer<>nil then FTerrainRenderer.OnGetTerrainBounds:=GetTerrainBounds;
  end;//else
end;

procedure tGLTiledRndLandscape.SetExtentX(const Value: integer);
begin
  FExtentX := Value;
  fExtentXhalf:=fExtentX div 2;
end;

procedure tGLTiledRndLandscape.SetExtentZ(const Value: integer);
begin
  FExtentZ := Value;
  fExtentZhalf:=fExtentZ div 2;
end;

procedure tGLTiledRndLandscape.SetGenerationRadius(const Value: integer);
var
  x,z,i	:integer;
begin
  FGenerationRadius := Value;
  SetLength(fGenRadius,sqr(FGenerationRadius*2+1));
  i:=0;
  for z:=0 to FGenerationRadius do begin
    for x:=1 to FGenerationRadius do begin
      if Trunc(sqrt(sqr(x)+sqr(z)))<=FGenerationRadius then begin
        fGenRadius[i].dx:=x; 	fGenRadius[i].dz:=z;
        fGenRadius[i+1].dx:=-z; 	fGenRadius[i+1].dz:=x;
        fGenRadius[i+2].dx:=-x; 	fGenRadius[i+2].dz:=-z;
        fGenRadius[i+3].dx:=z; 	fGenRadius[i+3].dz:=-x;
        Inc(i,4);
      end;//if
    end;//for
  end;//for
  SetLength(fGenRadius,i-3);
  fLandTiles.Capacity:=(i-3)*2;
end;

procedure tGLTiledRndLandscape.SetIsDefaultTile(const Value: tIsDefaultTile);
begin
  FIsDefaultTile:=Value;
end;

procedure tGLTiledRndLandscape.SetLandTileCapacity(const Value: integer);
begin
  FLandTileCapacity := Value;
end;

procedure tGLTiledRndLandscape.SetLandTileDensity(const Value: single);
begin
  FLandTileDensity := Value;
end;

procedure tGLTiledRndLandscape.SetOnCreateDefaultTile(
  const Value: TStartPreparingDataEvent);
begin
  FOnCreateDefaultTile := Value;
end;

procedure tGLTiledRndLandscape.SetOnCreateLandTile(const Value: tOnCreateLandTile);
begin
  FOnCreateLandTile:=Value;
end;

procedure tGLTiledRndLandscape.SetSeed(const Value: integer);
begin
  FSeed := Value;
  InitializeRandomGenerator(fSeed);
end;

procedure tGLTiledRndLandscape.SetSize(const aSize: integer);
begin
  fLandTileSize:=aSize;
end;

procedure tGLTiledRndLandscape.SetTerrainRenderer(const Value: TGLTerrainRenderer);
begin
  FTerrainRenderer := Value;
  FTerrainRenderer.HeightDataSource:=Self;
end;

procedure tGLTiledRndLandscape.StartPreparingData(HeightData: tHeightData);
var
  i		:integer;
  tx,tz	:integer;
begin
  with HeightData do begin
    DataState:=hdsPreparing;

    if (abs(XLeft)mod (HeightData.Size-1) =0) and
       (abs(YTop)mod (HeightData.Size-1) =0)
    then begin
      FindLandTile(XLeft,YTop,tx,tz);

      if IsDefaultTile(tx,tz) then begin
        OnCreateDefaultTile(HeightData);
        Exit;
      end;//if

      {Look if the landtile has already been computed}
      for i:=0 to fLandTiles.Count-1 do begin
        with tLandTile(fLandTiles.Items[i]).LandTileInfo do begin
          if (x=tx)and(z=tz) then begin
            if (State=hdsReady) then begin
              tLandTile(fLandTiles.Items[i]).StartPreparingData(HeightData);
              Exit;
              end
            else Break;
          end;//if
        end;//with
      end;//for
    end;//if

    DataState:=hdsNone;
  end;//with
end;

function tGLTiledRndLandscape.TileDistance(const x1, z1, x2,
  z2: integer): single;
begin
  Result:=sqrt(sqr(fExtentXhalf-abs(abs(x1-x2)-fExtentXhalf)) +
               sqr(fExtentZhalf-abs(abs(z1-z2)-fExtentZhalf)) );
end;

function tGLTiledRndLandscape.TileDistanceSquared(const x1,z1,x2,z2:integer): integer;
begin
  Result:=sqr(fExtentXhalf-abs(abs(x1-x2)-fExtentXhalf)) +
          sqr(fExtentZhalf-abs(abs(z1-z2)-fExtentZhalf));
end;

procedure tGLTiledRndLandscape.Update;
var
  i,j,maxi    :integer;
  maxd,d		:integer;
  cx,cz		:integer;
  cx0,cz0	:integer;
  Found		:boolean;
  NewLandTile	:tLandTile;
begin
  CameraPosition(cx0,cz0);
  if fMapUpdating or (fOldCamX=cx0)and(fOldCamZ=cz0) then exit;

  for j:=0 to High(fGenRadius) do begin
    fMapUpdating:=True;
    cx:=cx0+fGenRadius[j].dx;
    cz:=cz0+fGenRadius[j].dz;
    fIntegerConstrain(cx,cz);
    if IsDefaultTile(cx,cz) then continue;

    {Look if the landtile has already been computed}
    Found:=False;
    for i:=0 to fLandTiles.Count-1 do begin
      with tLandTile(fLandTiles.Items[i]).LandTileInfo do begin
        if (x=cx)and(z=cz)and(State=hdsReady) then begin
          Found:=True;
          Break;
        end;//if
      end;//with
    end;//for

    {If not, compute it}
    if not Found and not fLandTileComputing then begin
      if fLandTiles.Count>=FLandTileCapacity then begin // If the tile buffer is full...
        maxd:=-1;                                      // ...replace the farthest tile
        maxi:=-1;
        for i:=0 to fLandTiles.Count-1 do with tLandTile(fLandTiles.Items[i]) do begin
          d:=sqr(cx0-LandTileInfo.x)+sqr(cz0-LandTileInfo.z);
          if d>maxd then begin
            maxd:=d;
            maxi:=i;
          end;//if
        end;//for i
        if sqrt(maxd)>FGenerationRadius+1 then begin
          tLandTile(fLandTiles.Items[maxi]).Free;
        end;//if
      end;//if
      ComputeLandTile(cx,cz,NewLandTile);
      fMapUpdating:=False;
      Exit; // Don't explore further. Let it for the next time step
    end;//if
  end;//for j
  fMapUpdating:=False;
  fOldCamX:=cx0; // Surrounding completely updated, we can stop checking
  fOldCamZ:=cz0;
  fLandTiles.Pack;
end;

function tGLTiledRndLandscape.XMoveBoundary: single;
begin
  Result:=ExtentX*LandTileSize*0.95;
end;

function tGLTiledRndLandscape.ZMoveBoundary: single;
begin
  Result:=ExtentZ*LandTileSize*0.95;
end;

{ tGLFractalArchipelago }

procedure tGLFractalArchipelago.ComputeLandTile(const aX,aZ: integer;
                                               var NewLandTile:tLandTile);
begin
  NewLandTile:=tGLFractalHDS.CreateAsSlave(Self);
  inherited ComputeLandTile(aX,aZ,NewLandTile);
end;

constructor tGLFractalArchipelago.Create(AOwner: TComponent);
begin
  inherited;
  OnCreateLandTile:=fOnCreateLandTile;
  IsDefaultTile:=fIsDefaultTile;
  IslandDensity:=0.4;
  FWaveAmplitude:=2;
  FWaveSpeed:=20;
  Sea:=False; // Sea is drawn by the PostRender event
end;

procedure tGLFractalArchipelago.fOnCreateDefaultTile(heightData : THeightData);
var
  x,y		:Integer;
  rasterLine	:GLHeightData.PSingleArray;
  oldType 	:THeightDataType;
begin
  with HeightData do begin
    DataState:=hdsPreparing;
    oldType:=DataType;
    Allocate(hdtSingle);
    MaterialName:=fMaterialName;

    for y:=0 to HeightData.Size-1 do begin
      rasterLine:=singleRaster[y];
      for x:=0 to HeightData.Size-1 do begin
        rasterLine[x]:=fSeaLevel;
      end;//for
    end;//for

    if oldType<>hdtSingle then DataType:=oldType;
  end;//with
end;

procedure tGLFractalArchipelago.fOnCreateLandTile(aX,aZ,aSeed: integer;
  var aLandscape: tLandTile);
begin
  InitializeRandomGenerator(aSeed);
  with tGLFractalHDS(aLandscape) do begin
    {Initialize the tile}
    Seed:=random(MaxInt);
    Depth:=Self.FDepth;
    Amplitude:=random(fAmplitudeMax-fAmplitudeMin)+fAmplitudeMin;
    Roughness:=random*(fRoughnessMax-fRoughnessMin)+fRoughnessMin;
    ApplyLighting(aLandscape);
    ApplyTexture(aLandscape);
    ApplyTopography(aLandscape);
    Cyclic:=True;
    PrimerLandscape:=True;

    {Generate the landscape}
    PrimerIsland(SeaLevel-SeaTransparency,random*Amplitude/2,fHeight);  // Pre-generate an island
    BuildHeightField;
    if ErosionByRain.Enabled then DoErosionByRain;
    if ErosionByLife.Enabled then DoErosionByLife;
    if ErosionBySea.Enabled then DoErosionBySea;
    if Sea then DoSea;
    BuildNormals;
    if Lighting then BuildLightMap else ClearLightMap;
    BuildTexture;
    fNormal:=nil;
    fLightMap:=nil;
  end;//with
end;

procedure tGLFractalArchipelago.fPostRenderSeaDynamic(var rci: TRenderContextInfo;
  const heightDatas: TList);
{Code borrowed from Eric's Archipelago GLScene advanced demo}
var
   i, x, y, s, s2 : Integer;
   t : Single;
   hd : THeightData;
const
  r = 0.75;
  g = 0.75;
  b = 1;

  function WaterPhase(const px, py : Single) : Single;
  begin
     Result:=t*1+px*0.16+py*0.09;
  end;

  procedure IssuePoint(rx, ry : Integer);
  var
    px, py : Single;
    alpha, colorRatio, ca, sa : Single;
  begin
    px:=x+rx+s2;
    py:=y+ry+s2;
    if hd.DataState=hdsNone then begin
       alpha:=1;
    end else begin
      alpha:=(fSeaLevel-hd.SmallIntHeight(rx, ry))*(1/fSeaTransparency);
      alpha:=ClampValue(alpha, 0.5, 1);
    end;
    SinCos(WaterPhase(px, py)*fWaveSpeed, sa, ca);
    colorRatio:=1-alpha*0.1;
    glColor4f(r*colorRatio, g*colorRatio, b, alpha);
    glTexCoord2f(px*0.01+0.002*sa, py*0.01+0.0022*ca-t*0.01);
    glVertex3f(px, py, fSeaLevel+fWaveAmplitude*sa*VSF);
  end;

  begin
  //if not WaterPlane then Exit;
  t:=((GetTickCount-rhdsStartTime)/10000);
  FTerrainRenderer.MaterialLibrary.ApplyMaterial(fSeaMaterialName, rci);
  repeat
    //if not WasAboveWater then InvertGLFrontFace;
    glPushAttrib(GL_ENABLE_BIT);

    glDisable(GL_LIGHTING);
    glDisable(GL_NORMALIZE);

    glStencilFunc(GL_ALWAYS, 1, 255);
    glStencilMask(255);
    glStencilOp(GL_KEEP, GL_KEEP, GL_REPLACE);
    glEnable(GL_STENCIL_TEST);
    glNormal3f(0, 0, 1);

    for i:=0 to heightDatas.Count-1 do begin
       hd:=THeightData(heightDatas.List[i]);
       if (hd.DataState=hdsReady) and (hd.HeightMin>fSeaLevel) then continue;
       x:=hd.XLeft;
       y:=hd.YTop;
       s:=hd.Size-1;
       s2:=s div 2;
       glBegin(GL_TRIANGLE_FAN);
          IssuePoint(s2, s2);
          IssuePoint(0, 0);
          IssuePoint(s2, 0);
          IssuePoint(s, 0);
          IssuePoint(s, s2);
          IssuePoint(s, s);
          IssuePoint(s2, s);
          IssuePoint(0, s);
          IssuePoint(0, s2);
          IssuePoint(0, 0);
       glEnd;
    end;
    glStencilOp(GL_KEEP, GL_KEEP, GL_KEEP);
    glPopAttrib;
    //if not WasAboveWater then InvertGLFrontFace;
    //WaterPolyCount:=heightDatas.Count*8;
  until not FTerrainRenderer.MaterialLibrary.UnApplyMaterial(rci);
end;

procedure tGLFractalArchipelago.fPostRenderSeaStatic(var rci: TRenderContextInfo;
  const heightDatas: TList);
var
   i, x, y, s, s2 : Integer;
   hd 	: THeightData;
   t	:single;
const
  r = 0.75;
  g = 0.75;
  b = 1;

  procedure IssuePoint(rx, ry : Integer);
  var
    px, py 			: Single;
    alpha, colorRatio 	: Single;
  begin
    px:=x+rx+s2;
    py:=y+ry+s2;
    if hd.DataState=hdsNone then begin
       alpha:=1;
    end else begin
       alpha:=(fSeaLevel-hd.SmallIntHeight(rx, ry))*(1/fSeaTransparency);
       alpha:=ClampValue(alpha, 0.5, 1);
    end;
    colorRatio:=1-alpha*0.1;
    glColor4f(r*colorRatio, g*colorRatio, b, alpha);
    glTexCoord2f(px*0.01, py*0.01+t);
    glVertex3f(px, py, fSeaLevel);
  end;

  begin
  t:=Frac(GetTickCount/1000);

  FTerrainRenderer.MaterialLibrary.ApplyMaterial(fSeaMaterialName, rci);
  repeat
    //if not WasAboveWater then InvertGLFrontFace;
    glPushAttrib(GL_ENABLE_BIT);

    glDisable(GL_LIGHTING);
    glDisable(GL_NORMALIZE);

    glStencilFunc(GL_ALWAYS, 1, 255);
    glStencilMask(255);
    glStencilOp(GL_KEEP, GL_KEEP, GL_REPLACE);
    glEnable(GL_STENCIL_TEST);
    glNormal3f(0, 0, 1);

    for i:=0 to heightDatas.Count-1 do begin
       hd:=THeightData(heightDatas.List[i]);
       if (hd.DataState=hdsReady) and (hd.HeightMin>fSeaLevel) then continue;
       x:=hd.XLeft;
       y:=hd.YTop;
       s:=hd.Size-1;
       s2:=s div 2;
       glBegin(GL_TRIANGLE_FAN);
          IssuePoint(s2, s2);
          IssuePoint(0, 0);
          IssuePoint(s2, 0);
          IssuePoint(s, 0);
          IssuePoint(s, s2);
          IssuePoint(s, s);
          IssuePoint(s2, s);
          IssuePoint(0, s);
          IssuePoint(0, s2);
          IssuePoint(0, 0);
       glEnd;
    end;
    glStencilOp(GL_KEEP, GL_KEEP, GL_KEEP);
    glPopAttrib;
    //if not WasAboveWater then InvertGLFrontFace;
    //WaterPolyCount:=heightDatas.Count*8;
  until not FTerrainRenderer.MaterialLibrary.UnApplyMaterial(rci);
end;

function tGLFractalArchipelago.GetIslandDensity: single;
begin
  Result:=FLandTileDensity;
end;

procedure tGLFractalArchipelago.SetAmplitudeMax(const Value: integer);
begin
  FAmplitudeMax := Value;
end;

procedure tGLFractalArchipelago.SetAmplitudeMin(const Value: integer);
begin
  FAmplitudeMin := Value;
end;

procedure tGLFractalArchipelago.SetDepth(const Value: integer);
begin
  FDepth := Value;
  SetSize(Round(IntPower(2,fDepth)));
end;

procedure tGLFractalArchipelago.SetIslandDensity(const Value:single);
begin
  LandTileDensity := Value;
end;

procedure tGLFractalArchipelago.SetRoughnessMax(const Value: single);
begin
  FRoughnessMax := Value;
end;

procedure tGLFractalArchipelago.SetRoughnessMin(const Value: single);
begin
  FRoughnessMin := Value;
end;

procedure tGLFractalArchipelago.SetSeaDynamic(const Value: boolean);
begin
  FSeaDynamic := Value;
  if fSeaDynamic
  then FTerrainRenderer.OnHeightDataPostRender:=fPostRenderSeaDynamic
  else FTerrainRenderer.OnHeightDataPostRender:=fPostRenderSeaStatic;
end;

procedure tGLFractalArchipelago.SetSeaMaterialName(const Value: string);
begin
  FSeaMaterialName := Value;
end;

procedure tGLFractalArchipelago.SetTerrainRenderer(
  const Value: TGLTerrainRenderer);
begin
  inherited;
  SeaDynamic:=FSeaDynamic;  // Called to hook the PostRender event handler
end;

procedure tGLFractalArchipelago.SetWaveAmplitude(const Value: single);
begin
  FWaveAmplitude := Value;
end;

procedure tGLFractalArchipelago.SetWaveSpeed(const Value: single);
begin
  FWaveSpeed := Value;
end;

{***************************************************************}
{*******              RANDOM HDS ALGORITHMS             ********}
{***************************************************************}

procedure FractalMiddlePointHDS(const aDepth, aSeed,aAmplitude: integer;
                             const aRoughness:single; aCyclic:boolean;
                             var z:tMapOfSingle; var MinZ,MaxZ:single);
{Fractal algorithm based on the middle-point displacement method. It is built in
 a way that it can be juxtaposed seamlessly to itself (cyclic boundaries)}
var
  iter,stp,stp2	:integer;
  i,j			:integer;
  dz			:single;
  Size			:integer;

  procedure Let(var Z:single; const Value:single);
  {Fill variables only if they have not been predefined}
  begin
    if Z=empty then Z:=Value;
  end;

  function Get(const x,y:integer;var Value:single):boolean;
  {Fill variables only if they have not been predefined}
  begin
    Value:=z[x,y];
    Result:=(Value=empty);
  end;

  function Centre(const x,y,stp:integer):single;
  begin
    Result:=z[x-stp,y-stp];
    Result:=Result+z[x-stp,y+stp];
    Result:=Result+z[x+stp,y-stp];
    Result:=Result+z[x+stp,y+stp];
    Result:=Result*0.25;
    if MinZ>Result then MinZ:=Result;
    if MaxZ<Result then MaxZ:=Result;
  end;

  function Side(const x,y,stp:integer):single;
  var
    n	:integer;
  begin
    n:=0;
    Result:=0;
    if y-stp>=0 then begin
      Result:=Result+z[x,y-stp];
      Inc(n);
    end;
    if y+stp<=Size then begin
      Result:=Result+z[x,y+stp];
      Inc(n);
    end;
    if x-stp>=0 then begin
      Result:=Result+z[x-stp,y];
      Inc(n);
    end;
    if x+stp<=Size then begin
      Result:=Result+z[x+stp,y];
      Inc(n);
    end;
    Result:=Result/n;
    if MinZ>Result then MinZ:=Result;
    if MaxZ<Result then MaxZ:=Result;
  end;

begin
  InitializeRandomGenerator(aSeed);
  Size:=High(z);
  dz:=aAmplitude*VSF;
  MinZ:=1e38;
  MaxZ:=-1e38;

  if aCyclic then begin
    Let(z[0,0],0);
    Let(z[0,Size],z[0,0]);
    Let(z[Size,0],z[0,0]);
    Let(z[Size,Size],z[0,0]);

    {Build Height field}
    FOR iter:= 1 TO aDepth do begin         // iterations
      stp:=Round(Size/IntPower(2,(iter-1))); // step
      stp2:=stp div 2;                      // half step
      dz:=dz*aRoughness;
      i:=stp2;
      repeat
        j:=stp2;
        repeat // Centre
          if z[i,j]=Empty then begin
            z[i,j]:=Centre(i,j,stp2);
            z[i,j]:=z[i,j]+(random*dz*2-dz)*1.4;
          end;//if
          Inc(j,stp);
        until j>Size-stp2+1;
        Inc(i,stp);
      until i>Size-stp2+1;
      i:=stp2;
      repeat
        j:=0;
        repeat // Sides
          if z[i,j]=Empty then begin
            z[i,j]:=Side(i,j,stp2);
            z[i,j]:=z[i,j]+random*dz*2-dz;
          end;//if
          if z[j,i]=Empty then begin
            z[j,i]:=Side(j,i,stp2);
            z[j,i]:=z[j,i]+random*dz*2-dz;
          end;//if
          Inc(j,stp);
        until j>=Size;
        Let(z[Size,i],z[0,i]);
        Let(z[i,Size],z[i,0]);
        Inc(i,stp);
      until i>Size-stp2+1;
    end;//for iter
  end // if Cyclic


  else begin // Non-cyclic landscape
    Let(z[0,0],random*dz*2-dz);
    Let(z[0,Size],random*dz*2-dz);
    Let(z[Size,0],random*dz*2-dz);
    Let(z[Size,Size],random*dz*2-dz);

    {Build Height field}
    FOR iter:= 1 TO aDepth do begin         // iterations
      stp:=Round(Size/IntPower(2,(iter-1))); // step
      stp2:=stp div 2;                      // half step
      dz:=dz*aRoughness;
      i:=stp2;
      repeat
        j:=stp2;
        repeat // Centre
          if z[i,j]=Empty then begin
            z[i,j]:=Centre(i,j,stp2);
            z[i,j]:=z[i,j]+(random*dz*2-dz)*1.4;
          end;
          Inc(j,stp);
        until j>Size-stp2+1;
        Inc(i,stp);
      until i>Size-stp2+1;
      i:=stp2;
      repeat
        j:=0;
        repeat // Sides
          if z[i,j]=Empty then begin
            z[i,j]:=Side(i,j,stp2);
            z[i,j]:=z[i,j]+random*dz*2-dz;
          end;//if
          if z[j,i]=Empty then begin
            z[j,i]:=Side(j,i,stp2);
            z[j,i]:=z[j,i]+random*dz*2-dz;
          end;//if
          Inc(j,stp);
        until j>Size;
        Inc(i,stp);
      until i>Size-stp2+1;
    end;//for iter
  end;//else Cyclic
end;

{***************************************************************}
{*******              PREDEFINED HEIGHT-FIELD           ********}
{***************************************************************}

procedure PrimerNull(var z:tMapOfSingle);
{Empty field}
var
  x,y	:integer;
  Size	:integer;
begin
  Size:=High(z);
  for y:=0 to Size do begin
    for x:=0 to Size do begin
      z[x,y]:=Empty;
    end;//for
  end;//for
end;

procedure PrimerIsland(LowZ,HighZ:single;var z:tMapOfSingle);
{Ensure that the border of the tile is low (below sea level) and the middle
is high.}
var
  i		:integer;
  Size	:integer;
begin
  Size:=High(z);
  PrimerNull(z);
  HighZ:=HighZ*VSF;
  LowZ:=LowZ*VSF;
  z[Size div 2,Size div 2]:=HighZ;
  for i:=0 to Size do begin
    z[i,0]:=LowZ;
    z[0,i]:=LowZ;
    z[Size,i]:=LowZ;
    z[i,Size]:=LowZ;
  end;//for i
end;

Initialization
  rhdsStartTime:=GetTickCount;

end.

